/*-------------------------------------------------------
   CODESET.C -- Telnet Client, Terminal Emulator Program
             R[hϊ
   Copyright (c) 2002-2005
        Hiroto Shigeta. All rights reserved.
  -------------------------------------------------------*/
#include <windows.h>
#include "monitor.h"
#include "socket.h"

#define SS2      142
#define ISEUC(A) (A >= 161 && A <= 254)

#define IS_UTF8_2(A) (A >= 0xC0 && A <= 0xDF)
#define IS_UTF8_3(A) (A >= 0xE0 && A <= 0xEF)

/* R[hZbg(0=EUC, 1=Shift-JIS, 2=UTF-8) */
int iCodeSet = 0;

/* JIS(LinuxΉ) */
BOOL bJisKanji = FALSE;

void jis2sjis(int *p1, int *p2)
{
	unsigned char c1         = *p1;
	unsigned char c2         = *p2;
	int           rowOffset  = c1 < 95 ? 112 : 176;
	int           cellOffset = c1 % 2 ? (c2 > 95 ? 32 : 31) : 126;

	*p1 = ((c1 + 1) >> 1) + rowOffset;
	*p2 += cellOffset;
}

void sjis2jis(int *p1, int *p2)
{
	unsigned char c1         = *p1;
	unsigned char c2         = *p2;
	int           adjust     = c2 < 159;
	int           rowOffset  = c1 < 160 ? 112 : 176;
	int           cellOffset = adjust ? (c2 > 127 ? 32 : 31) : 126;

	*p1 = ((c1 - rowOffset) << 1) - adjust;
	*p2 -= cellOffset;
}

void jis2euc(int *p1, int *p2)
{
	*p1 += 128;
	*p2 += 128;
}

void euc2jis(int *p1, int *p2)
{
	*p1 -= 128;
	*p2 -= 128;
}

void euc2sjis(unsigned char *c1, unsigned char *c2)
{
	int p1 = (int)*c1;
	int p2 = (int)*c2;

	euc2jis(&p1, &p2);
	jis2sjis(&p1, &p2);

	*c1 = (unsigned char)p1;
	*c2 = (unsigned char)p2;
}

void sjis2euc(unsigned char *c1, unsigned char *c2)
{
	int p1 = (int)*c1;
	int p2 = (int)*c2;

	sjis2jis(&p1, &p2);
	jis2euc(&p1, &p2);

	*c1 = (unsigned char)p1;
	*c2 = (unsigned char)p2;
}

int utf8tosjis(unsigned char *c1, unsigned char *c2, unsigned char *c3)
{
	wchar_t       wch;
	int           ret;
	unsigned char out[2];

	if (IS_UTF8_2(*c1)) {
		wch = *c1 & 0x1F;
		wch = wch << 6;
		wch |= *c2 & 0x3F;
	} else {
		wch = *c1 & 0x0F;
		wch = wch << 6;
		wch |= *c2 & 0x3F;
		wch = wch << 6;
		wch |= *c3 & 0x3F;
	}
	ret = wctomb(out, wch);
	if (ret > 0) {
		*c1 = out[0];
		if (ret > 1) {
			*c2 = out[1];
		}
	}
	return ret;
}

int sjis2utf8(unsigned char *c1, unsigned char *c2, unsigned char *c3)
{
	unsigned char in[2];
	int           ret;
	wchar_t       wch;

	in[0] = *c1;
	in[1] = *c2;
	ret = mbtowc(&wch, in, *c2 == 0 ? 1 : 2);
	if (ret < 1) {
		return ret;
	}
	if (wch <= 0x07FF) {
		*c1 = (unsigned char)(wch >> 6) | 0xC0;
		*c2 = (unsigned char)(wch & 0x003F) | 0x80;
		ret = 2;
	} else {
		*c1 = (unsigned char)(wch >> 12) | 0xE0;
		*c2 = (unsigned char)(wch >> 6 & 0x003F) | 0x80;
		*c3 = (unsigned char)(wch & 0x003F) | 0x80;
		ret = 3;
	}
	return ret;
}

/* pJ^JiM */
int SendToServerKatakana(char chCharCode)
{
	char chCharCode2, chCharCode3;

	switch (iCodeSet) {
	case 1:// Shift-JIS
		SendToServerChar(chCharCode);
		break;
	case 2:// UTF-8
		sjis2utf8(&chCharCode, &chCharCode2, &chCharCode3);
		SendToServerChar(chCharCode);
		SendToServerChar(chCharCode2);
		SendToServerChar(chCharCode3);
		break;
	default:// EUC
		SendToServerChar((char)SS2);
		SendToServerChar(chCharCode);
		break;
	}
	return 0;
}

/* (2Byte)M */
int SendToServerKanji(char chCharCode1, char chCharCode2)
{
	char chCharCode3;
	int  nBytes = 0;

	switch (iCodeSet) {
	case 1:// Shift-JIS
		break;
	case 2:// UTF-8
		nBytes = sjis2utf8(&chCharCode2, &chCharCode1, &chCharCode3);
		break;
	default:// EUC
		sjis2euc(&chCharCode2, &chCharCode1);
		break;
	}
	SendToServerChar(chCharCode2);
	SendToServerChar(chCharCode1);
	if (nBytes > 2) {
		SendToServerChar(chCharCode3);
	}
	return 0;
}

/* Byte̐擪Byte? */
BOOL IsHostCodeSetLeadByte(char chCharCode)
{
	BOOL bRet;

	if (bJisKanji) {
		return TRUE;
	}
	switch (iCodeSet) {
	case 1:// Shift-JIS
		bRet = IsDBCSLeadByte(chCharCode);
		break;
	case 2:// UTF-8
		bRet = IS_UTF8_2((int)(chCharCode & 0xFF)) || IS_UTF8_3((int)(chCharCode & 0xFF));
		break;
	default:// EUC
		bRet = ISEUC((int)(chCharCode & 0xFF)) || (int)(chCharCode & 0xFF) == SS2;
		break;
	}
	return bRet;
}

/* ʂɊ(2Byte)o */
int MonOutKanji(char chCharCode1, char chCharCode2)
{
	static char chSecondByte = 0;
	int         nBytes;

	if (bJisKanji) {
		int n1 = (int)(chCharCode1 & 0xFF);
		int n2 = (int)(chCharCode2 & 0xFF);

		jis2sjis(&n1, &n2);
		MonOutTwo((char)n1, (char)n2);
		DebugRcvPrint("%c%c", (char)n1, (char)n2);
		return 0;
	}
	switch (iCodeSet) {
	case 1:// Shift-JIS
		MonOutTwo(chCharCode1, chCharCode2);
		DebugRcvPrint("%c%c", chCharCode1, chCharCode2);
		break;
	case 2:// UTF-8
		if (IS_UTF8_3((int)(chCharCode1 & 0xFF)) && chSecondByte == 0) {
			chSecondByte = chCharCode2;
			return 1;
		}
		if (IS_UTF8_2((int)(chCharCode1 & 0xFF))) {
			chSecondByte = chCharCode2;
			chCharCode2 = 0;
		}
		nBytes = utf8tosjis(&chCharCode1, &chSecondByte, &chCharCode2);
		switch (nBytes) {
		case 1:
			MonOutOne(chCharCode1);
			DebugRcvPrint("%c", chCharCode1);
			break;
		case 2:
			MonOutTwo(chCharCode1, chSecondByte);
			DebugRcvPrint("%c%c", chCharCode1, chSecondByte);
			break;
		default:
			if ((chCharCode1 & 0xFF) == 0xE3 &&
				(chSecondByte & 0xFF) == 0x82 &&
				(chCharCode2 & 0xFF) == 0x99) {
				VoicedSoundMark(FALSE);
			} else if ((chCharCode1 & 0xFF) == 0xE3 &&
					   (chSecondByte & 0xFF) == 0x82 &&
					   (chCharCode2 & 0xFF) == 0x9A) {
				VoicedSoundMark(TRUE);
			} else {
				MonOutOne('?');
			}
			DebugRcvPrint("\\%d", (int)(chCharCode1 & 0xFF));
			DebugRcvPrint("\\%d", (int)(chSecondByte & 0xFF));
			if (chCharCode2 != 0) {
				DebugRcvPrint("\\%d", (int)(chCharCode2 & 0xFF));
			}
			break;
		}
		chSecondByte = 0;
		break;
	default:// EUC
		if ((int)(chCharCode1 & 0xFF) == SS2) {
			MonOutOne(chCharCode2);
			DebugRcvPrint("%c", chCharCode2);
		} else {
			euc2sjis(&chCharCode1, &chCharCode2);
			MonOutTwo(chCharCode1, chCharCode2);
			DebugRcvPrint("%c%c", chCharCode1, chCharCode2);
		}
		break;
	}
	return 0;
}

/* JIS7J^Jiϊ */
BOOL IsShiftCode(unsigned char *c)
{
	static BOOL bSO = FALSE;

	switch (*c) {
	case 0x0E:
	case 0x0F:
		bSO = *c == 0x0E;
		return TRUE;
	default:
		if (bSO && *c >= 0x21 && *c <= 0x5F) {
			*c += 0x80;
		}
		break;
	}
	return FALSE;
}

/* obt@̕R[hϊ */
char *ConvertCodeSet(char *pszBuffer)
{
	size_t       nBufferLength;
	char         *pszConvertBuffer = NULL;
	unsigned int i = 0;
	unsigned int j = 0;
	int          nBytes;

	switch (iCodeSet) {
	case 1:// Shift-JIS
		return pszBuffer;
	case 2:// UTF-8
		nBufferLength = strlen(pszBuffer);
		pszConvertBuffer = malloc(nBufferLength * 3 + 1);
		if (pszConvertBuffer == NULL) {
			return NULL;
		}
		memset(pszConvertBuffer, 0, nBufferLength * 3 + 1);
		while (pszBuffer[i] != 0) {
			if (IsDBCSLeadByte(pszBuffer[i])) {
				pszConvertBuffer[j]     = pszBuffer[i];
				pszConvertBuffer[j + 1] = pszBuffer[i + 1];
				nBytes = sjis2utf8(&pszConvertBuffer[j],
								   &pszConvertBuffer[j + 1],
								   &pszConvertBuffer[j + 2]);
				i += 2;
				j += nBytes;
			} else if (pszBuffer[i] >= '' && pszBuffer[i] <= '') {
				pszConvertBuffer[j] = pszBuffer[i];
				nBytes = sjis2utf8(&pszConvertBuffer[j],
								   &pszConvertBuffer[j + 1],
								   &pszConvertBuffer[j + 2]);
				i++;
				j += nBytes;
			} else if (pszBuffer[i] == '\n') {
				i++;
			} else {
				pszConvertBuffer[j] = pszBuffer[i];
				i++;
				j++;
			}
		}
		break;
	default:// EUC
		nBufferLength = strlen(pszBuffer);
		pszConvertBuffer = malloc(nBufferLength * 2 + 1);
		if (pszConvertBuffer == NULL) {
			return NULL;
		}
		memset(pszConvertBuffer, 0, nBufferLength * 2 + 1);
		while (pszBuffer[i] != 0) {
			if (IsDBCSLeadByte(pszBuffer[i])) {
				pszConvertBuffer[j]     = pszBuffer[i];
				pszConvertBuffer[j + 1] = pszBuffer[i + 1];
				sjis2euc(&pszConvertBuffer[j], &pszConvertBuffer[j + 1]);
				i += 2;
				j += 2;
			} else if (pszBuffer[i] >= '' && pszBuffer[i] <= '') {
				pszConvertBuffer[j]     = (char)SS2;
				pszConvertBuffer[j + 1] = pszBuffer[i];
				i++;
				j += 2;
			} else if (pszBuffer[i] == '\n') {
				i++;
			} else {
				pszConvertBuffer[j] = pszBuffer[i];
				i++;
				j++;
			}
		}
		break;
	}
	free(pszBuffer);
	return pszConvertBuffer;
}
