/*-------------------------------------------------------
   MONITOR.C -- Telnet Client, Terminal Emulator Program
             EBhE
   Copyright (c) 2002-2006
        Hiroto Shigeta. All rights reserved.
  -------------------------------------------------------*/
#include <winsock2.h>
#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <limits.h>
#include <locale.h>
#include "codeset.h"
#include "socket.h"
#include "terminal.h"
#include "mycolor.h"
#include "resource.h"

#ifndef WM_MOUSEWHEEL
#define WM_MOUSEWHEEL                   0x020A
#endif /* WM_MOUSEWHEEL */
#ifndef WHEEL_DELTA
#define WHEEL_DELTA                     120     /* Value for rolling one detent */
#endif /* WHEEL_DELTA */
#ifndef WHEEL_PAGESCROLL
#define WHEEL_PAGESCROLL                (UINT_MAX) /* Scroll one page */
#endif /* WHEEL_PAGESCROLL */
#ifndef SPI_GETWHEELSCROLLLINES
#define SPI_GETWHEELSCROLLLINES   104
#endif /* SPI_GETWHEELSCROLLLINES */

/* Variables in CODESET.C */
extern BOOL bJisKanji;

/* Variables in TELNET.C */
extern int Echo;
extern int EnvNewOld;

/* Functions in TELNET.C */
int RecvDataCheck(char in_c);
int SendNAWS(void);

/* Functions in MISC.C */
void GetAfterSpaceCutString(const char *pszSrc, UINT cchLen, char *pszDst, UINT cchSize);
void SetSelArea(int xSelFrom, int ySelFrom, int xSelTo, int ySelTo, RECT *prc);
BOOL IsDBCSSecondByte(const char *pstrString, int iIndex);
int GetWordPos(const char *pszLine, int iPos, int *piWordFrom, int *piWordTo);
int GetFontCharSize(HWND hwnd, HFONT hFont, int *pcxChar, int *pcyChar);

/* Functions in REGISTRY.C */
BOOL RegReadFont(PLOGFONT pLogFont);
BOOL RegWriteFont(PLOGFONT pLogFont);
BOOL RegReadTermSize(int *piCols, int *piRows, UINT *pnBackLogMax);
BOOL RegWriteTermSize(int iCols, int iRows, UINT nBackLogMax);
BOOL RegReadAutoClose(BOOL *pbAutoClose);
BOOL RegWriteAutoClose(BOOL bAutoClose);
BOOL InitMRU(HWND hCombo);
int  FindMRU(const char *pszHostName, int *piCodeSet, char *pszTermType, size_t cchTermType);
BOOL AddMRU(const char *pszHostName);
BOOL DelMRU(int iIndex);

/* sobt@\ */
struct tagLINEBUF {
	struct tagLINEBUF *pPrevLine;
	struct tagLINEBUF *pNextLine;
	char              *pszLine;
	char              *pchAttr;
	int               iLen;
};

typedef struct tagLINEBUF LINEBUF, *PLINEBUF;

static char szAppName[] = "SigTerm";

static char szHost[64];

static HWND      hwnd;
static HINSTANCE hInst;
static int       x = 0;
static int       y = 0;
static int       cxChar;
static int       cyChar;
static int       iSpace;				/* Ԋul */

static SCROLLINFO vsi;
static BOOL bThumbTrack = FALSE;

static int nCOLS = 80;
static int nROWS = 24;

static PLINEBUF pTopLine     = NULL;
static PLINEBUF pEndLine     = NULL;
static PLINEBUF pViewTopLine = NULL;
static int      iViewTopLine = 0;
static PLINEBUF pMonTopLine  = NULL;
static UINT     nLineCount   = 0;
static UINT     nBackLogMax  = INT_MAX;

static int iScrollFrom;
static int iScrollTo;
static int xSaveCaret;
static int ySaveCaret;

static char chAttribute;				/* \ */

static int xSelectFrom = 0;
static int ySelectFrom = 0;
static int xSelectTo   = 0;
static int ySelectTo   = 0;

static BOOL bRightEnd = FALSE;

static BOOL bHideCaret = TRUE;
static int  iCaretSize = 1;

static BOOL bAutoClose = FALSE;

/* vg^Cv */
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

/* \ݒ(ǉ) */
void AddAttr(char chAttr)
{
	chAttribute |= chAttr;
}

/* \NA */
void ClearAttr(void)
{
	chAttribute = 0;
}

/* Ï̃[W쐬 */
HRGN CreateSelRgn(RECT *prc, int iHeadAdj, int iTailAdj)
{
	HRGN hRgn1, hRgn2;

	if (prc->top != prc->bottom) {
		hRgn1 = CreateRectRgn((prc->left + iHeadAdj) * cxChar, prc->top * cyChar,
							  cxChar * nCOLS, (prc->top + 1) * cyChar);
		hRgn2 = CreateRectRgn(0, prc->bottom * cyChar,
							  (prc->right + iTailAdj) * cxChar, (prc->bottom + 1) * cyChar);
		CombineRgn(hRgn1, hRgn1, hRgn2, RGN_OR);
		DeleteObject(hRgn2);
		if (prc->top + 1 != prc->bottom) {
			hRgn2 = CreateRectRgn(0, (prc->top + 1) * cyChar,
								  cxChar * nCOLS, prc->bottom * cyChar);
			CombineRgn(hRgn1, hRgn1, hRgn2, RGN_OR);
			DeleteObject(hRgn2);
		}
	} else {
		hRgn1 = CreateRectRgn((prc->left + iHeadAdj) * cxChar, prc->top * cyChar,
							  (prc->right + iTailAdj) * cxChar, (prc->top + 1) * cyChar);
	}
	return hRgn1;
}

/* Ï𖳌(ĕ`) */
void InvalidateSelArea(HWND hwnd, RECT *prc, int iOffset)
{
	HRGN hRgn;

	hRgn = CreateSelRgn(prc, 0, 1);
	OffsetRgn(hRgn, 0, iOffset);
	InvalidateRgn(hwnd, hRgn, TRUE);
	DeleteObject(hRgn);
}

/* Ï̍𖳌(ĕ`) */
void InvalidateSelChanged(HWND hwnd, RECT *prc1, RECT *prc2, int iOffset)
{
	HRGN hRgn1, hRgn2;
	int iHeadAdj1 = 0;
	int iHeadAdj2 = 0;
	int iTailAdj1 = 0;
	int iTailAdj2 = 0;

	if (prc1->top == prc2->top && prc1->left == prc2->left) {
		if (prc1->bottom == prc2->bottom) {
			if (prc1->right < prc2->right) {
				iTailAdj2 = 1;
			} else if (prc1->right > prc2->right) {
				iTailAdj1 = 1;
			}
		} else if (prc1->bottom < prc2->bottom) {
			iTailAdj2 = 1;
		} else {
			iTailAdj1 = 1;
		}
	} else if (prc1->top == prc2->top && prc1->bottom == prc2->bottom) {
		if (prc1->left == prc1->right && prc2->left != prc2->right) {
			iTailAdj2 = 1;
		} else if (prc2->left == prc2->right && prc1->left != prc1->right) {
			iTailAdj1 = 1;
		} else if (prc1->left > prc2->left) {
			iHeadAdj1 = 1;
		} else if (prc1->left < prc2->left) {
			iHeadAdj2 = 1;
		}
	} else if (prc1->top == prc2->top + 1 && prc1->bottom == prc2->bottom &&
			   prc1->right != prc2->right) {
		iTailAdj1 = 1;
	} else if (prc1->top + 1 == prc2->top && prc1->bottom == prc2->bottom &&
			   prc1->right != prc2->right) {
		iTailAdj2 = 1;
	}
	hRgn1 = CreateSelRgn(prc1, iHeadAdj1, iTailAdj1);
	hRgn2 = CreateSelRgn(prc2, iHeadAdj2, iTailAdj2);
	CombineRgn(hRgn1, hRgn1, hRgn2, RGN_XOR);
	DeleteObject(hRgn2);
	OffsetRgn(hRgn1, 0, iOffset);
	InvalidateRgn(hwnd, hRgn1, TRUE);
	DeleteObject(hRgn1);
}

/* Lbg̐ݒ */
void SetCaret(void)
{
	PLINEBUF pLineBuf;
	int      i;
	int      ypos;
	BOOL     bDBCS;

	if (hwnd != GetFocus()) {
		return;
	}
	ypos = y - vsi.nPos + (vsi.nMax - max((int)vsi.nPage - 1, 0));
	if (ypos > nROWS - 1) {
		if (!bHideCaret) {
			HideCaret(hwnd);
			bHideCaret = TRUE;
		}
	} else {
		pLineBuf = pMonTopLine;
		for (i = 0; i < y; i++) {
			pLineBuf = pLineBuf->pNextLine;
		}
		bDBCS = IsDBCSLeadByte(pLineBuf->pszLine[x]);
		if ((bDBCS && iCaretSize != 2) || (!bDBCS && iCaretSize != 1)) {
			if (!bHideCaret) {
				HideCaret(hwnd);
				bHideCaret = TRUE;
			}
			DestroyCaret();
			iCaretSize = bDBCS ? 2 : 1;
			CreateCaret(hwnd, NULL, cxChar * iCaretSize, cyChar);
		}
		if (bHideCaret) {
			ShowCaret(hwnd);
			bHideCaret = FALSE;
		}
		SetCaretPos(x * cxChar, ypos * cyChar);
	}
}

/* LbgʒuL */
void SaveCaret(void)
{
	xSaveCaret = x;
	ySaveCaret = y;
}

/* Lbgʒu𕜌 */
void LoadCaret(void)
{
	x = xSaveCaret;
	y = ySaveCaret;
	SetCaret();
}

/* sobt@̉ */
void FreeLineBuf(PLINEBUF pLineBuf)
{
	free(pLineBuf->pszLine);
	free(pLineBuf->pchAttr);
	free(pLineBuf);
	nLineCount--;
}

/* sobt@̍쐬 */
PLINEBUF CreateNewLine(PLINEBUF pPrevLine, PLINEBUF pNextLine)
{
	PLINEBUF pLineBuf;

	if (nLineCount == INT_MAX) {
		return NULL;
	}
	pLineBuf = (PLINEBUF)malloc(sizeof(LINEBUF));
	if (pLineBuf == NULL) {
		return NULL;
	}
	pLineBuf->pPrevLine = pPrevLine;
	pLineBuf->pNextLine = pNextLine;
	pLineBuf->pszLine = malloc(nCOLS + 1);
	if (pLineBuf->pszLine == NULL) {
		free(pLineBuf);
		return NULL;
	}
	memset(pLineBuf->pszLine, ' ', nCOLS);
	pLineBuf->pszLine[nCOLS] = 0;
	pLineBuf->pchAttr = malloc(nCOLS);
	if (pLineBuf->pchAttr == NULL) {
		free(pLineBuf->pszLine);
		free(pLineBuf);
		return NULL;
	}
	memset(pLineBuf->pchAttr, 0, nCOLS);
	pLineBuf->iLen = nCOLS;
	nLineCount++;

	return pLineBuf;
}

/* ʂNAăLbgz[|WVɖ߂ */
void ClearMon(void)
{
	PLINEBUF pLineBuf;

	pLineBuf = pMonTopLine;
	while (pLineBuf != NULL) {
		memset(pLineBuf->pszLine, ' ', nCOLS);
		memset(pLineBuf->pchAttr, 0, nCOLS);
		pLineBuf = pLineBuf->pNextLine;
	}
	InvalidateRect(hwnd, NULL, TRUE);
	x = y = 0;
	SetCaret();
}

/* Lbg荶̕\NA */
void ClearMonBeforeCaret(void)
{
	PLINEBUF pLineBuf;
	int      i;
	RECT     rect;

	pLineBuf = pMonTopLine;
	for (i = 0; i < y; i++) {
		memset(pLineBuf->pszLine, ' ', nCOLS);
		memset(pLineBuf->pchAttr, 0, nCOLS);
		pLineBuf = pLineBuf->pNextLine;
	}
	SetRect(&rect, 0, 0, nCOLS * cxChar, y * cyChar);
	InvalidateRect(hwnd, &rect, TRUE);
	if (x > 0) {
		memset(pLineBuf->pszLine, ' ', x);
		memset(pLineBuf->pchAttr, 0, x);
		SetRect(&rect, 0, y * cyChar, x * cxChar, (y + 1) * cyChar);
		InvalidateRect(hwnd, &rect, TRUE);
	}
}

/* Lbg肠Ƃ̕\NA */
/* Lbgʒu܂(HP-UXΉ)      */
void ClearMonAfterCaret(void)
{
	PLINEBUF pLineBuf;
	int      i;
	RECT     rect;

	pLineBuf = pMonTopLine;
	for (i = 0; i < y; i++) {
		pLineBuf = pLineBuf->pNextLine;
	}
	memset(&pLineBuf->pszLine[x], ' ', nCOLS - x);
	memset(&pLineBuf->pchAttr[x], 0, nCOLS - x);
	SetRect(&rect, x * cxChar, y * cyChar, nCOLS * cxChar, (y + 1) * cyChar);
	InvalidateRect(hwnd, &rect, TRUE);

	pLineBuf = pLineBuf->pNextLine;
	while (pLineBuf != NULL) {
		memset(pLineBuf->pszLine, ' ', nCOLS);
		memset(pLineBuf->pchAttr, 0, nCOLS);
		pLineBuf = pLineBuf->pNextLine;
	}
	SetRect(&rect, 0, (y + 1) * cyChar, nCOLS * cxChar, nROWS * cyChar);
	InvalidateRect(hwnd, &rect, TRUE);
	SetCaret();
}

/* s̐擪Lbg̈ʒu܂łNA */
void ClearLineBeforeCaret(void)
{
	PLINEBUF pLineBuf;
	int      i;
	RECT     rect;

	pLineBuf = pMonTopLine;
	for (i = 0; i < y; i++) {
		pLineBuf = pLineBuf->pNextLine;
	}
	memset(pLineBuf->pszLine, ' ', x);
	memset(pLineBuf->pchAttr, 0, x);
	SetRect(&rect, 0, y * cyChar, x * cxChar, (y + 1) * cyChar);
	InvalidateRect(hwnd, &rect, TRUE);
}

/* Lbg̈ʒus̏I܂łNA */
void ClearLineAfterCaret(void)
{
	PLINEBUF pLineBuf;
	int      i;
	RECT     rect;

	pLineBuf = pMonTopLine;
	for (i = 0; i < y; i++) {
		pLineBuf = pLineBuf->pNextLine;
	}
	if (IsDBCSSecondByte(pLineBuf->pszLine, x)) {
		pLineBuf->pszLine[x - 1] = ' ';
		pLineBuf->pchAttr[x - 1] = 0;
	}
	memset(&pLineBuf->pszLine[x], ' ', nCOLS - x);
	memset(&pLineBuf->pchAttr[x], 0, nCOLS - x);
	SetRect(&rect, x * cxChar, y * cyChar, nCOLS * cxChar, (y + 1) * cyChar);
	InvalidateRect(hwnd, &rect, TRUE);
	SetCaret();
}

/* ʂ1sփXN[ */
void ScrollUpMon(void)
{
	PLINEBUF pFromLine, pToLine;
	int      i;
	RECT     rect;

	SendMessage(hwnd, WM_VSCROLL, SB_BOTTOM, 0L);
	pFromLine = pMonTopLine;
	for (i = 0; i < iScrollFrom; i++) {
		pFromLine = pFromLine->pNextLine;
	}
	pToLine = pFromLine;
	for (i = 0; i < iScrollTo - iScrollFrom; i++) {
		pToLine = pToLine->pNextLine;
	}
	if (iScrollFrom == 0) {
		if (pTopLine == pMonTopLine) {
			pTopLine = pToLine;
		}
		pMonTopLine = pToLine;
		pViewTopLine = pToLine;
	}
	if (iScrollTo == nROWS - 1) {
		pEndLine = pEndLine->pPrevLine;
	}
	if (pToLine->pNextLine != NULL) {
		pToLine->pNextLine->pPrevLine = pToLine->pPrevLine;
	}
	pToLine->pPrevLine->pNextLine = pToLine->pNextLine;
	pToLine->pNextLine = pFromLine;
	pToLine->pPrevLine = pFromLine->pPrevLine;
	if (pFromLine->pPrevLine != NULL) {
		pFromLine->pPrevLine->pNextLine = pToLine;
	}
	pFromLine->pPrevLine = pToLine;
	memset(pToLine->pszLine, ' ', nCOLS);
	memset(pToLine->pchAttr, 0, nCOLS);
	SetRect(&rect, 0, iScrollFrom * cyChar, cxChar * nCOLS, (iScrollTo + 1) * cyChar);
	ScrollWindowEx(hwnd, 0, cyChar, &rect, &rect, NULL, NULL, SW_ERASE | SW_INVALIDATE);
}

/* ʂ1sփXN[ */
void ScrollDownMon(void)
{
	PLINEBUF pFromLine, pToLine;
	int      i;
	RECT     rect;

	SendMessage(hwnd, WM_VSCROLL, SB_BOTTOM, 0L);
	pFromLine = pMonTopLine;
	for (i = 0; i < iScrollFrom; i++) {
		pFromLine = pFromLine->pNextLine;
	}
	pToLine = pFromLine;
	for (i = 0; i < iScrollTo - iScrollFrom; i++) {
		pToLine = pToLine->pNextLine;
	}
	if (iScrollFrom == 0) {
		if (pTopLine == pMonTopLine) {
			pTopLine = pTopLine->pNextLine;
		}
		pMonTopLine = pMonTopLine->pNextLine;
		pViewTopLine = pViewTopLine->pNextLine;
	}
	if (iScrollTo == nROWS - 1) {
		pEndLine = pFromLine;
	}
	if (pFromLine->pPrevLine != NULL) {
		pFromLine->pPrevLine->pNextLine = pFromLine->pNextLine;
	}
	pFromLine->pNextLine->pPrevLine = pFromLine->pPrevLine;
	pFromLine->pPrevLine = pToLine;
	pFromLine->pNextLine = pToLine->pNextLine;
	if (pToLine->pNextLine != NULL) {
		pToLine->pNextLine->pPrevLine = pFromLine;
	}
	pToLine->pNextLine = pFromLine;
	memset(pFromLine->pszLine, ' ', nCOLS);
	memset(pFromLine->pchAttr, 0, nCOLS);
	SetRect(&rect, 0, iScrollFrom * cyChar, cxChar * nCOLS, (iScrollTo + 1) * cyChar);
	ScrollWindowEx(hwnd, 0, -cyChar, &rect, &rect, NULL, NULL, SW_ERASE | SW_INVALIDATE);
}

/* XN[͈͂̐ݒ */
void SetScrollRangeMon(int iFrom, int iTo)
{
	iScrollFrom = iFrom;
	iScrollTo   = iTo;
}

/* LbgʒủɎw肳ꂽ̍sǉ */
/* Lbgʒȕ...(HP-UX)                 */
void AddLine(int iNum)
{
	int iSaveScrollFrom, iSaveScrollTo, i;

	iSaveScrollFrom = iScrollFrom;
	iSaveScrollTo   = iScrollTo;
	iScrollFrom = y;
	iScrollTo   = nROWS - 1;
	for (i = 0; i < iNum; i++) {
		ScrollUpMon();
	}
	iScrollFrom = iSaveScrollFrom;
	iScrollTo   = iSaveScrollTo;
}

/* Lbgs牺̍sw肳ꂽ폜 */
void DeleteLine(int iNum)
{
	int iSaveScrollFrom, iSaveScrollTo, i;

	iSaveScrollFrom = iScrollFrom;
	iSaveScrollTo   = iScrollTo;
	iScrollFrom = y;
	iScrollTo   = nROWS - 1;
	for (i = 0; i < iNum; i++) {
		ScrollDownMon();
	}
	iScrollFrom = iSaveScrollFrom;
	iScrollTo   = iSaveScrollTo;
}

/* Lbg̈ʒuw肳ꂽ̕폜 */
void DeleteChar(int iNum)
{
	PLINEBUF pLineBuf;
	int      i;
	RECT     rect;

	pLineBuf = pMonTopLine;
	for (i = 0; i < y; i++) {
		pLineBuf = pLineBuf->pNextLine;
	}
	i = x;
	while (pLineBuf->pszLine[i + iNum] != 0) {
		pLineBuf->pszLine[i] = pLineBuf->pszLine[i + iNum];
		pLineBuf->pchAttr[i] = pLineBuf->pchAttr[i + iNum];
		i++;
	}
	memset(&pLineBuf->pszLine[nCOLS - 1 - iNum], ' ', iNum);
	memset(&pLineBuf->pchAttr[nCOLS - 1 - iNum], 0, iNum);
	SetRect(&rect, x * cxChar, y * cyChar, cxChar * nCOLS, (y + 1) * cyChar);
	ScrollWindowEx(hwnd, -cxChar * iNum, 0, &rect, &rect, NULL, NULL,
				   SW_ERASE | SW_INVALIDATE);
}

/* LbgWwňړ */
void MoveCaret(int xpos, int ypos)
{
	x = xpos;
	y = ypos;
	SetCaret();
}

/* Lbgֈړ */
void MoveCaretUpOne(void)
{
	y--;
	SetCaret();
}

/* Lbgֈړ */
void MoveCaretDownOne(void)
{
	PLINEBUF pLineBuf = NULL;
	RECT     rect;
	int      iScrollInc = 0;

	if (iScrollTo < nROWS - 1 && y == iScrollTo) {
		ScrollDownMon();
	} else {
		y++;
	}
	if (y >= nROWS) {
		if (iScrollFrom > 0) {
			ScrollDownMon();
		} else {
			if (!bThumbTrack) {
				SendMessage(hwnd, WM_VSCROLL, SB_BOTTOM, 0L);
			}
			if (nLineCount < nBackLogMax + (UINT)nROWS) {
				pLineBuf = CreateNewLine(pEndLine, NULL);
			}
			if (pLineBuf == NULL) {// obNOėp
				while (pTopLine->iLen != nCOLS) {
					if (pViewTopLine == pTopLine) {
						pViewTopLine = pViewTopLine->pNextLine;
						iScrollInc++;
					} else {
						iViewTopLine--;
						vsi.nPos--;
					}
					pTopLine = pTopLine->pNextLine;
					FreeLineBuf(pTopLine->pPrevLine);
					pTopLine->pPrevLine = NULL;
					vsi.fMask = SIF_POS | SIF_RANGE | SIF_DISABLENOSCROLL;
					vsi.nMax--;
					SetScrollInfo(hwnd, SB_VERT, &vsi, TRUE);
					if (ySelectFrom == 0) {
						if (ySelectTo == 0) {
							xSelectFrom = ySelectFrom = xSelectTo = ySelectTo = 0;
						} else {
							xSelectFrom = 0;
							ySelectTo--;
						}
					} else if (ySelectFrom > 0) {
						ySelectFrom--;
						ySelectTo--;
					}
				}
				if (pViewTopLine == pMonTopLine || pViewTopLine == pTopLine) {
					pViewTopLine = pViewTopLine->pNextLine;
					iScrollInc++;
				} else {
					iViewTopLine--;
					vsi.fMask = SIF_POS;
					vsi.nPos--;
					SetScrollInfo(hwnd, SB_VERT, &vsi, TRUE);
				}
				pMonTopLine = pMonTopLine->pNextLine;
				pLineBuf = pTopLine;
				pTopLine = pTopLine->pNextLine;
				pTopLine->pPrevLine = NULL;
				pLineBuf->pPrevLine = pEndLine;
				pLineBuf->pNextLine = NULL;
				memset(pLineBuf->pszLine, ' ', nCOLS);
				memset(pLineBuf->pchAttr, 0, nCOLS);
				pEndLine->pNextLine = pLineBuf;
				pEndLine = pLineBuf;
				if (ySelectFrom == 0) {
					if (ySelectTo == 0) {
						xSelectFrom = ySelectFrom = xSelectTo = ySelectTo = 0;
					} else {
						xSelectFrom = 0;
						ySelectTo--;
					}
				} else if (ySelectFrom > 0) {
					ySelectFrom--;
					ySelectTo--;
				}
			} else {
				pEndLine->pNextLine = pLineBuf;
				pEndLine = pLineBuf;
				if (pViewTopLine == pMonTopLine) {
					pViewTopLine = pViewTopLine->pNextLine;
					iViewTopLine++;
					vsi.nPos++;
					iScrollInc = 1;
				}
				pMonTopLine = pMonTopLine->pNextLine;
				vsi.fMask = SIF_POS | SIF_RANGE | SIF_DISABLENOSCROLL;
				vsi.nMax++;
				SetScrollInfo(hwnd, SB_VERT, &vsi, TRUE);
			}
			if (iScrollInc > 0) {
				SetRect(&rect, 0, 0, cxChar * nCOLS, cyChar * nROWS);
				ScrollWindowEx(hwnd, 0, -cyChar * iScrollInc, &rect, &rect, NULL, NULL,
							   SW_ERASE | SW_INVALIDATE);
			}
		}
		y = nROWS - 1;
	}
	SetCaret();
}

/* Lbg[ֈړ */
void MoveCaretLeftTop(void)
{
	x = 0;
	SetCaret();
}

/* Lbgֈړ */
void MoveCaretLeftOne(void)
{
	if (x > 0) {
		if (bRightEnd) {
			bRightEnd = FALSE;
		} else {
			x--;
		}
	} else if (y > 0) {
		y--;
		x = nCOLS - 1;
	}
	SetCaret();
}

/* LbgEֈړ */
void MoveCaretRightOne(void)
{
	x++;
	SetCaret();
}

/* Lbg̃^uֈړ */
void MoveCaretNextTab(void)
{
	x = min(x + (8 - x % 8), nCOLS - 1);
	SetCaret();
}

/* ʂ1oCgo */
int MonOutOne(char c)
{
	PLINEBUF pLineBuf;
	int      i;
	RECT     rect;
	int      ypos;

	if (!bThumbTrack) {
		SendMessage(hwnd, WM_VSCROLL, SB_BOTTOM, 0L);
	}
	if (x == nCOLS - 1 && bRightEnd) {
		MoveCaretLeftTop();
		MoveCaretDownOne();
		bRightEnd = FALSE;
	}
	pLineBuf = pMonTopLine;
	for (i = 0; i < y; i++) {
		pLineBuf = pLineBuf->pNextLine;
	}
	if (IsDBCSSecondByte(pLineBuf->pszLine, x)) {
		pLineBuf->pszLine[x - 1] = ' ';
		pLineBuf->pchAttr[x - 1] = 0;
	}
	pLineBuf->pszLine[x] = c;
	pLineBuf->pchAttr[x] = chAttribute;
	ypos = y - vsi.nPos + (vsi.nMax - max((int)vsi.nPage - 1, 0));
	SetRect(&rect, x * cxChar, ypos * cyChar, (x + 2) * cxChar, (ypos + 1) * cyChar);
	InvalidateRect(hwnd, &rect, TRUE);
	UpdateWindow(hwnd);
	if (x == nCOLS - 1) {
		bRightEnd = TRUE;
	} else {
		x++;
		bRightEnd = FALSE;
	}
	SetCaret();

	return 0;
}

/* ʂ2oCgo */
int MonOutTwo(char c1, char c2)
{
	PLINEBUF pLineBuf;
	int      i;
	RECT     rect;
	int      ypos;

	if (!bThumbTrack) {
		SendMessage(hwnd, WM_VSCROLL, SB_BOTTOM, 0L);
	}
	if (x == nCOLS - 1) {
		MoveCaretLeftTop();
		MoveCaretDownOne();
		bRightEnd = FALSE;
	}
	pLineBuf = pMonTopLine;
	for (i = 0; i < y; i++) {
		pLineBuf = pLineBuf->pNextLine;
	}
	pLineBuf->pszLine[x]     = c1;
	pLineBuf->pchAttr[x]     = chAttribute;
	pLineBuf->pszLine[x + 1] = c2;
	pLineBuf->pchAttr[x + 1] = chAttribute;
	ypos = y - vsi.nPos + (vsi.nMax - max((int)vsi.nPage - 1, 0));
	SetRect(&rect, x * cxChar, ypos * cyChar, (x + 3) * cxChar, (ypos + 1) * cyChar);
	InvalidateRect(hwnd, &rect, TRUE);
	UpdateWindow(hwnd);
	if (x == nCOLS - 2) {
		x++;
		bRightEnd = TRUE;
	} else {
		x += 2;
		bRightEnd = FALSE;
	}
	SetCaret();

	return 0;
}

/* O̕ɑ_܂͔_t(Mac OS X) */
int VoicedSoundMark(BOOL bSemiVoicedSoundMark)
{
	PLINEBUF pLineBuf;
	int      i;
	int      xpos, ypos;
	RECT     rect;

	if (!bThumbTrack) {
		SendMessage(hwnd, WM_VSCROLL, SB_BOTTOM, 0L);
	}
	if (x < 2) {
		return 0;
	}
	pLineBuf = pMonTopLine;
	for (i = 0; i < y; i++) {
		pLineBuf = pLineBuf->pNextLine;
	}
	xpos = x - (bRightEnd ? 1 : 2);
	if (*(LPWORD)&pLineBuf->pszLine[xpos] == 0x4583) {	/* E(gGfBA) */
		pLineBuf->pszLine[xpos + 1] += 79;				/* E   */
	} else {
		pLineBuf->pszLine[xpos + 1] += bSemiVoicedSoundMark ? 2 : 1;
	}
	ypos = y - vsi.nPos + (vsi.nMax - max((int)vsi.nPage - 1, 0));
	SetRect(&rect, xpos * cxChar, ypos * cyChar, (xpos + 2) * cxChar, (ypos + 1) * cyChar);
	InvalidateRect(hwnd, &rect, TRUE);
	UpdateWindow(hwnd);
	return 0;
}

/* ʃTCY擾 */
int GetTerminalWindowSize(short *rows, short *cols)
{
	*rows = nROWS;
	*cols = nCOLS;

	return 0;
}

/* ʂ̃TCY */
BOOL ResizeMon(int iOldCols, int iOldRows)
{
	int      i, iCopySize;
	PLINEBUF pLineBuf, pEnd, pMon;
	PLINEBUF pTop = NULL;

	pMon = pMonTopLine;
	iCopySize = min(nCOLS, iOldCols);
	for (i = 0; i < nROWS; i++) {
		pLineBuf = CreateNewLine(NULL, NULL);
		if (pLineBuf == NULL) {
			/* 肩̍sobt@ */
			while (pTop != NULL) {
				pLineBuf = pTop->pNextLine;
				FreeLineBuf(pTop);
				pTop = pLineBuf;
			}
			return FALSE;
		}
		if (pTop == NULL) {
			pLineBuf->pPrevLine = pMonTopLine->pPrevLine;
			pTop = pLineBuf;
			pEnd = pLineBuf;
		} else {
			pLineBuf->pPrevLine = pEnd;
			pEnd->pNextLine = pLineBuf;
			pEnd = pLineBuf;
		}
		if (pMon != NULL) {
			memcpy(pLineBuf->pszLine, pMon->pszLine, iCopySize);
			memcpy(pLineBuf->pchAttr, pMon->pchAttr, iCopySize);
			pMon = pMon->pNextLine;
		}
	}
	if (pMonTopLine->pPrevLine != NULL) {
		pMonTopLine->pPrevLine->pNextLine = pTop;
	}
	if (pTopLine == pMonTopLine) {
		pTopLine = pTop;
	}
	if (pViewTopLine == pMonTopLine) {
		pViewTopLine = pTop;
	}
	/* Âsobt@ */
	while (pMonTopLine != NULL) {
		pLineBuf = pMonTopLine->pNextLine;
		FreeLineBuf(pMonTopLine);
		pMonTopLine = pLineBuf;
	}
	pMonTopLine = pTop;
	pEndLine = pEnd;
	return TRUE;
}

/* LvV̏ */
void FormatCaption(HWND hwnd, ...)
{
	HINSTANCE hInstance;
	va_list   marker;
	char      szFormat[32];
	char      szTitle[128];

	hInstance = (HINSTANCE)GetWindowLong(hwnd, GWL_HINSTANCE);
	LoadString(hInstance, IDS_CAPTION, szFormat, sizeof(szFormat));
	va_start(marker, hwnd);
	FormatMessage(FORMAT_MESSAGE_FROM_STRING,
				  szFormat, 0, 0, szTitle, sizeof(szTitle), &marker);
	va_end(marker);
	SetWindowText(hwnd, szTitle);
}

/* ڑs */
void ConnectErrorMessage(HWND hwnd, DWORD dwErrCode)
{
	char szString[1024];
	char szFormat[64];
	int  id;
	char *pszHostName = szHost;

	if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
					  NULL,
					  dwErrCode,
					  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
					  szString,
					  sizeof(szString),
					  NULL) == 0) {
		LoadString(hInst, IDS_CONNECTNG, szString, sizeof(szString));
	}
	MessageBox(hwnd, szString, szAppName, MB_OK | MB_ICONEXCLAMATION);
	id = FindMRU(szHost, NULL, NULL, 0);
	if (id != -1) {
		LoadString(hInst, IDS_DELMRU, szFormat, sizeof(szFormat));
		FormatMessage(FORMAT_MESSAGE_FROM_STRING | FORMAT_MESSAGE_ARGUMENT_ARRAY,
					  szFormat, 0, 0, szString, sizeof(szString), (va_list *)&pszHostName);
		if (MessageBox(hwnd, szString, szAppName, MB_YESNO | MB_ICONQUESTION) == IDYES) {
			DelMRU(id);
			memset(szHost, 0, sizeof(szHost));
		}
	}
	LoadString(hInst, IDS_DISCONNECTED, szString, sizeof(szString));
	FormatCaption(hwnd, szAppName, szString);
}

/* R}hC̉ */
void ReadArguments(HWND hwnd, char *pszCmdLine)
{
	char          cOption = '\0';
	char          *pOptStart = NULL;
	unsigned long nOptLen;

	while (*pszCmdLine != '\0' || pOptStart != NULL) {
		switch (*pszCmdLine) {
		case '/': case ' ': case '\t': case '\0':
			if (pOptStart != NULL) {
				nOptLen = pszCmdLine - pOptStart;
				switch (cOption) {
				case 'H':
					if (nOptLen < sizeof(szHost)) {
						memset(szHost, 0, sizeof(szHost));
						strncpy(szHost, pOptStart, nOptLen);
					}
					break;
				case 'C':
					if (nOptLen == 4 && _strnicmp(pOptStart, "SJIS", 4) == 0) {
						iCodeSet = 1;
					} else if (nOptLen == 4 && _strnicmp(pOptStart, "UTF8", 4) == 0) {
						iCodeSet = 2;
					} else if (nOptLen == 3 && _strnicmp(pOptStart, "EUC", 3) == 0) {
						iCodeSet = 0;
					}
					break;
				case 'T':
					if (nOptLen == 4 && _strnicmp(pOptStart, "vt52", 4) == 0) {
						strcpy(szTerminalType, "vt52");
					} else if (nOptLen == 5 && _strnicmp(pOptStart, "vt100", 5) == 0) {
						strcpy(szTerminalType, "vt100");
					}
					break;
				}
				cOption = '\0';
				pOptStart = NULL;
			}
			if (*pszCmdLine == '/') {
				pszCmdLine++;
				switch (*pszCmdLine) {
				case 'H': case 'C': case 'T':
					cOption = *pszCmdLine;
					break;
				case 'h': case 'c': case 't':
					cOption = *pszCmdLine - 0x20;
					break;
				default:
					cOption = '\0';
					break;
				}
			}
			break;
		default:
			if (cOption != '\0' && pOptStart == NULL) {
				pOptStart = pszCmdLine;
			}
			break;
		}
		if (*pszCmdLine != '\0') {
			pszCmdLine++;
		}
	}
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR pszCmdLine, int iCmdShow)
{
	WNDCLASSEX wcex;
	MSG        msg;
	WSADATA    wsaData;
	HACCEL     hAccel;
	int        rc;
	char       szString[32];

	setlocale(LC_ALL, "japanese");
	if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0) {
		return 1;
	}
	hInst = hInstance;
	RegReadTermSize(&nCOLS, &nROWS, &nBackLogMax);
	RegReadAutoClose(&bAutoClose);
	ReadArguments(hwnd, pszCmdLine);

	wcex.cbSize        = sizeof(WNDCLASSEX);
	wcex.style         = CS_DBLCLKS;
	wcex.lpfnWndProc   = WndProc;
	wcex.cbClsExtra    = 0;
	wcex.cbWndExtra    = 0;
	wcex.hInstance     = hInstance;
	wcex.hIcon         = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_SIGTERM));
	wcex.hCursor       = LoadCursor(NULL, IDC_IBEAM);
	wcex.hbrBackground = GetMyBkBrush();
	wcex.lpszMenuName  = MAKEINTRESOURCE(IDR_MAINMENU);
	wcex.lpszClassName = szAppName;
	wcex.hIconSm       = LoadImage(hInstance, MAKEINTRESOURCE(IDI_SIGTERM),
									   IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
	RegisterClassEx(&wcex);

	hwnd = CreateWindowEx(WS_EX_CLIENTEDGE, szAppName, szAppName,
						  WS_OVERLAPPEDWINDOW | WS_VSCROLL,
						  CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);
	if (hwnd == NULL) {
		WSACleanup();
		return 1;
	}
	/* EBhETCY̒ */
	SetWindowPos(hwnd, NULL, 0, 0,
				 cxChar * nCOLS + GetSystemMetrics(SM_CXSIZEFRAME) * 2 +
				 GetSystemMetrics(SM_CXEDGE) * 2 +
				 GetSystemMetrics(SM_CXVSCROLL),
				 cyChar * nROWS + GetSystemMetrics(SM_CYSIZEFRAME) * 2 +
				 GetSystemMetrics(SM_CYEDGE) * 2 +
				 GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU),
				 SWP_NOMOVE | SWP_NOZORDER | SWP_NOREDRAW);
	ShowWindow(hwnd, iCmdShow);
	UpdateWindow(hwnd);

	hAccel = LoadAccelerators(hInstance, MAKEINTRESOURCE(IDR_ACCELERATOR));

	if (szHost[0] != '\0') {
		LoadString(hInst, IDS_CONNECTING, szString, sizeof(szString));
		FormatCaption(hwnd, szAppName, szString);
		rc = ConnectServer(hwnd, szHost);
		if (rc != 0) {
			ConnectErrorMessage(hwnd, rc);
		}
	}
	while (GetMessage(&msg, NULL, 0, 0)) {
		if (!TranslateAccelerator(hwnd, hAccel, &msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	WSACleanup();
	return msg.wParam;
}

/* ڑ */
BOOL CALLBACK ConnectDlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	char szSelHost[64];
	int  iSelCodeSet;
	char szSelTermType[32];
	int  id = -1;
	BOOL bBeforeAutoClose;

	switch (uMsg) {
	case WM_INITDIALOG:
		CheckDlgButton(hwndDlg, IDC_AUTOCLOSE, bAutoClose);
		InitMRU(GetDlgItem(hwndDlg, IDC_HOST));
		if (SendMessage(GetDlgItem(hwndDlg, IDC_HOST), CB_GETCOUNT, 0, 0) > 0) {
			if (szHost[0] == '\0') {
				SendMessage(GetDlgItem(hwndDlg, IDC_HOST), CB_SETCURSEL, 0, 0);
				GetDlgItemText(hwndDlg, IDC_HOST, szSelHost, sizeof(szSelHost));
			} else {
				strcpy(szSelHost, szHost);
			}
			id = FindMRU(szSelHost, &iSelCodeSet, szSelTermType, sizeof(szSelTermType));
			if (id != -1) {
				SendMessage(GetDlgItem(hwndDlg, IDC_HOST), CB_SETCURSEL, id, 0);
				CheckRadioButton(hwndDlg, IDC_CODESET0, IDC_CODESET2,
								 IDC_CODESET0 + iSelCodeSet);
				CheckRadioButton(hwndDlg, IDC_TERMTYPE0, IDC_TERMTYPE1,
								 szSelTermType[2] == '5' ? IDC_TERMTYPE0 : IDC_TERMTYPE1);
				return TRUE;
			}
		}
		if (id == -1 && szHost[0] != '\0') {
			SetDlgItemText(hwndDlg, IDC_HOST, szHost);
		}
		CheckRadioButton(hwndDlg, IDC_CODESET0, IDC_CODESET2, IDC_CODESET0 + iCodeSet);
		CheckRadioButton(hwndDlg, IDC_TERMTYPE0, IDC_TERMTYPE1,
						 szTerminalType[2] == '5' ? IDC_TERMTYPE0 : IDC_TERMTYPE1);
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case IDC_HOST:
			switch (HIWORD(wParam)) {
			case CBN_SELCHANGE:
				SendMessage((HWND)lParam, CB_GETLBTEXT,
							SendMessage((HWND)lParam, CB_GETCURSEL, 0, 0), (LPARAM)szSelHost);
				id = FindMRU(szSelHost, &iSelCodeSet, szSelTermType, sizeof(szSelTermType));
				if (id != -1) {
					CheckRadioButton(hwndDlg, IDC_CODESET0, IDC_CODESET2,
									 IDC_CODESET0 + iSelCodeSet);
					CheckRadioButton(hwndDlg, IDC_TERMTYPE0, IDC_TERMTYPE1,
									 szSelTermType[2] == '5' ? IDC_TERMTYPE0 : IDC_TERMTYPE1);
				}
				break;
			}
			return TRUE;
		case IDOK:
			GetDlgItemText(hwndDlg, IDC_HOST, szHost, sizeof(szHost));
			if (IsDlgButtonChecked(hwndDlg, IDC_CODESET1)) {
				iCodeSet = 1;
			} else if (IsDlgButtonChecked(hwndDlg, IDC_CODESET2)) {
				iCodeSet = 2;
			} else {
				iCodeSet = 0;
			}
			if (IsDlgButtonChecked(hwndDlg, IDC_TERMTYPE0)) {
				strcpy(szTerminalType, "vt52");
			} else {
				strcpy(szTerminalType, "vt100");
			}
			bBeforeAutoClose = bAutoClose;
			bAutoClose = IsDlgButtonChecked(hwndDlg, IDC_AUTOCLOSE);
			if (bAutoClose != bBeforeAutoClose) {
				RegWriteAutoClose(bAutoClose);
			}
			EndDialog(hwndDlg, 1);
			return TRUE;
		case IDCANCEL:
			EndDialog(hwndDlg, 0);
			return TRUE;
		}
		break;
	}
	return FALSE;
}

/* tHg̑I */
BOOL OpenChooseFontDlg(HWND hwnd, PLOGFONT pLogFont)
{
	CHOOSEFONT cf;

	cf.lStructSize    = sizeof(CHOOSEFONT);
	cf.hwndOwner      = hwnd;
	cf.hDC            = NULL;
	cf.lpLogFont      = pLogFont;
	cf.iPointSize     = 0;
	cf.Flags          = CF_INITTOLOGFONTSTRUCT | CF_SCREENFONTS | CF_FIXEDPITCHONLY;
	cf.rgbColors      = 0L;
	cf.lCustData      = 0L;
	cf.lpfnHook       = NULL;
	cf.lpTemplateName = NULL;
	cf.hInstance      = NULL;
	cf.lpszStyle      = NULL;
	cf.nFontType      = 0;
	cf.nSizeMin       = 0;
	cf.nSizeMax       = 0;

	return ChooseFont(&cf);
}

/* ^[~iTCY */
BOOL CALLBACK TermSizeDlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	int iOldCols, iOldRows;

	switch (uMsg) {
	case WM_INITDIALOG:
		SetDlgItemInt(hwndDlg, IDC_COLS, nCOLS, FALSE);
		SetDlgItemInt(hwndDlg, IDC_ROWS, nROWS, FALSE);
		SendMessage(GetDlgItem(hwndDlg, IDC_COLSSPIN), UDM_SETRANGE, 0,
					MAKELONG(163, 40));
		SendMessage(GetDlgItem(hwndDlg, IDC_ROWSSPIN), UDM_SETRANGE, 0,
					MAKELONG(107, 12));
		SendMessage(GetDlgItem(hwndDlg, IDC_COLS), EM_LIMITTEXT, 3, 0L);
		SendMessage(GetDlgItem(hwndDlg, IDC_ROWS), EM_LIMITTEXT, 3, 0L);
		SetDlgItemInt(hwndDlg, IDC_BACKLOGMAX, nBackLogMax, FALSE);
		SendMessage(GetDlgItem(hwndDlg, IDC_BACKLOGMAXSPIN), UDM_SETRANGE32, 0, INT_MAX);
		SendMessage(GetDlgItem(hwndDlg, IDC_BACKLOGMAX), EM_LIMITTEXT, 10, 0L);
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case IDC_BACKLOGMAXMAX:
			SetDlgItemInt(hwndDlg, IDC_BACKLOGMAX, INT_MAX, FALSE);
			return TRUE;
		case IDOK:
			iOldCols = nCOLS;
			iOldRows = nROWS;
			nCOLS = GetDlgItemInt(hwndDlg, IDC_COLS, NULL, FALSE);
			nROWS = GetDlgItemInt(hwndDlg, IDC_ROWS, NULL, FALSE);
			nCOLS = max(40, min(nCOLS, 163));
			nROWS = max(12, min(nROWS, 107));
			if (ResizeMon(iOldCols, iOldRows)) {
				nBackLogMax = GetDlgItemInt(hwndDlg, IDC_BACKLOGMAX, NULL, FALSE);
				nBackLogMax = nBackLogMax < (UINT)INT_MAX ? nBackLogMax : INT_MAX;
				EndDialog(hwndDlg, 1);
				return TRUE;
			}/* TCYɎsƂ̓TCY߂0Ԃ */
			nCOLS = iOldCols;
			nROWS = iOldRows;
		case IDCANCEL:
			EndDialog(hwndDlg, 0);
			return TRUE;
		}
		break;
	}
	return FALSE;
}

/* o[W */
BOOL CALLBACK AboutDlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
	case WM_INITDIALOG:
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case IDOK:
		case IDCANCEL:
			EndDialog(hwndDlg, 0);
			return TRUE;
		}
		break;
	}
	return FALSE;
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	HDC             hdc;
	PAINTSTRUCT     ps;
	POINT           point;
	RECT            rect;
	int             rc;
	char            c;
	char            s[BUFSIZ];
	int             i;
	int             iPaintBeg;
	int             iPaintEnd;
	static LOGFONT  logfont;		/* tHg */
	static HFONT    hFont;			/* tHg */
	HFONT           hOldFont;		/* tHg([N) */
	HIMC            hIMC;			/* IMEp */
	COMPOSITIONFORM cf;				/* IMEp */
	char            chCharCode1;
	char            chCharCode2;
	PLINEBUF        pLineBuf, pPrevLine;
	int             iScrollInc;
	static BOOL     bLeftButtonDown = FALSE;
	UINT            nTotalLen;						/* Nbv{[hp */
	char            *pszCopyText, *pszPasteText;	/* Nbv{[hp */
	HGLOBAL         hClipText;						/* Nbv{[hp */
	PVOID           pClipText;						/* Nbv{[hp */
	static char     *pszPasteBuffer = NULL;
	static size_t   nPasteLength    = 0;
	static UINT     nPastePosition  = 0;
	char            szString[1024];
	static int      iWordFrom, iWordTo;				/* PI */
	static BOOL     bWordSelect = FALSE;			/* PI */
	static BOOL     bDiscClient = FALSE;			/* NCAgؒf */

	switch (message) {
	case WM_KEYDOWN:
		if (VTKeyDown(wParam, lParam)) {
			return 0;
		}
		break;
	case WM_CHAR:
		if (Soc != INVALID_SOCKET) {
			if (Echo) {
				MonOutOne((char)wParam);
			}
			if ((char)(wParam & 0xFF) >= '' && (char)(wParam & 0xFF) <= '') {
				SendToServerKatakana((char)wParam);
			} else {
				SendToServerChar((char)wParam);
			}
		}
		return 0;
	case WM_IME_CHAR:
		if (Soc != INVALID_SOCKET) {
			chCharCode1 = (char)(wParam & 0xFF);
			chCharCode2 = (char)(wParam >> 8);
			if (chCharCode2 == 0) {/* pJ^Ji */
				if (Echo) {
					MonOutOne(chCharCode1);
				}
				if ((char)(chCharCode1 & 0xFF) >= '' && (char)(chCharCode1 & 0xFF) <= '') {
					SendToServerKatakana(chCharCode1);
				} else {
					SendToServerChar(chCharCode1);
				}
			} else {
				if (Echo) {
					MonOutTwo(chCharCode2, chCharCode1);
				}
				SendToServerKanji(chCharCode1, chCharCode2);
			}
		}
		return 0;
	case SM_EVENT:
		switch (WSAGETSELECTEVENT(lParam)) {
		case FD_CONNECT:
			rc = WSAGETSELECTERROR(lParam);
			if (rc != 0) {
				closesocket(Soc);
				Soc = INVALID_SOCKET;
				ConnectErrorMessage(hwnd, rc);
				break;
			}
			FormatCaption(hwnd, szAppName, szHost);
			/* [̏Ԃ */
			iScrollFrom = 0;
			iScrollTo   = nROWS - 1;
			xSaveCaret  = 0;
			ySaveCaret  = 0;
			chAttribute = 0;
			memset(szString, 0, sizeof(szString));
			strcpy(szString, szHost);
			szString[strlen(szHost) + 1] = iCodeSet + '0';
			strcpy(szString + strlen(szHost) + 3, szTerminalType);
			AddMRU(szString);
			break;
		case FD_CLOSE:
			while (1) {
				rc = recv(Soc, &c, 1, 0);
				if (rc == SOCKET_ERROR || rc == 0) {
					break;
				}
				RecvDataCheck(c);
			}
			shutdown(Soc, SD_BOTH);
			closesocket(Soc);
			Soc = INVALID_SOCKET;
			LoadString(hInst, IDS_DISCONNECTED, szString, sizeof(szString));
			FormatCaption(hwnd, szAppName, szString);
			/* ʐM̏Ԃ */
			Echo = 1;
			EnvNewOld = -1;
			if (pszPasteBuffer != NULL) {
				free(pszPasteBuffer);
				pszPasteBuffer = NULL;
			}
			bJisKanji = FALSE;
			if (bDiscClient) {
				bDiscClient = FALSE;
			} else if (bAutoClose) {
				PostMessage(hwnd, WM_CLOSE, 0, 0L);
			}
			break;
		case FD_READ:
			rc = recv(Soc, s, sizeof(s), 0);
			if (rc != SOCKET_ERROR && rc != 0) {
				for (i = 0; i < rc; i++) {
					RecvDataCheck(s[i]);
				}
			}
			break;
		case FD_WRITE:
			if (pszPasteBuffer != NULL) {
				do {
					rc = send(Soc, &pszPasteBuffer[nPastePosition], nPasteLength, 0);
					if (rc == SOCKET_ERROR) {
						if (WSAGetLastError() == WSAEWOULDBLOCK) {
							return 0;
						}
						break;
					} else {
						nPasteLength   -= rc;
						nPastePosition += rc;
					}
				} while (nPasteLength > 0);
				free(pszPasteBuffer);
				pszPasteBuffer = NULL;
			}
			break;
		}
		return 0;
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case IDM_CONNECT:
			if (Soc != INVALID_SOCKET) {
				break;
			}
			if (!DialogBox(hInst, MAKEINTRESOURCE(IDD_CONNECT), hwnd, ConnectDlgProc)) {
				break;
			}
			LoadString(hInst, IDS_CONNECTING, szString, sizeof(szString));
			FormatCaption(hwnd, szAppName, szString);
			rc = ConnectServer(hwnd, szHost);
			if (rc != 0) {
				ConnectErrorMessage(hwnd, rc);
			}
			break;
		case IDM_DISCONNECT:
			if (Soc == INVALID_SOCKET) {
				break;
			}
			shutdown(Soc, SD_SEND);
			bDiscClient = TRUE;
			break;
		case IDM_EXIT:
			SendMessage(hwnd, WM_CLOSE, 0, 0L);
			break;
		case IDM_COPY:
			if (xSelectFrom == xSelectTo && ySelectFrom == ySelectTo) {
				return 0;
			}
			SetSelArea(xSelectFrom, ySelectFrom, xSelectTo, ySelectTo, &rect);
			pLineBuf = pTopLine;
			for (i = 0; i < rect.top; i++) {
				pLineBuf = pLineBuf->pNextLine;
			}
			if (rect.top == rect.bottom) {
				GetAfterSpaceCutString(&pLineBuf->pszLine[rect.left],
									   rect.right - rect.left,
									   szString,
									   sizeof(szString));
				nTotalLen = strlen(szString);
				pszCopyText = malloc(nTotalLen + 1);
				if (pszCopyText == NULL) {
					return -1;
				}
				strcpy(pszCopyText, szString);
			} else {
				GetAfterSpaceCutString(&pLineBuf->pszLine[rect.left],
									   strlen(&pLineBuf->pszLine[rect.left]),
									   szString,
									   sizeof(szString));
				nTotalLen = strlen(szString);
				pszCopyText = malloc(nTotalLen + 1);
				if (pszCopyText == NULL) {
					return -1;
				}
				strcpy(pszCopyText, szString);
				for (i = rect.top; i < rect.bottom - 1; i++) {
					pLineBuf = pLineBuf->pNextLine;
					GetAfterSpaceCutString(pLineBuf->pszLine,
										   strlen(pLineBuf->pszLine),
										   szString,
										   sizeof(szString));
					nTotalLen += strlen(szString) + 2;
					pszCopyText = realloc(pszCopyText, nTotalLen + 1);
					if (pszCopyText == NULL) {
						return -1;
					}
					strcat(pszCopyText, "\r\n");
					strcat(pszCopyText, szString);
				}
				pLineBuf = pLineBuf->pNextLine;
				GetAfterSpaceCutString(pLineBuf->pszLine,
									   rect.right,
									   szString,
									   sizeof(szString));
				nTotalLen += strlen(szString) + 2;
				pszCopyText = realloc(pszCopyText, nTotalLen + 1);
				if (pszCopyText == NULL) {
					return -1;
				}
				strcat(pszCopyText, "\r\n");
				strcat(pszCopyText, szString);
			}
			if (NULL == (hClipText = GlobalAlloc(GHND, nTotalLen + 1))) {
				free(pszCopyText);
				return 0;
			}
			pClipText = GlobalLock(hClipText);
			memcpy(pClipText, pszCopyText, nTotalLen);	// GHNDŃ[NAĂ
			GlobalUnlock(hClipText);
			free(pszCopyText);
			if (OpenClipboard(hwnd)) {
				EmptyClipboard();
				SetClipboardData(CF_TEXT, hClipText);
				CloseClipboard();
				return 0;
			}
			GlobalFree(hClipText);
			break;
		case IDM_PASTE:
			if (Soc == INVALID_SOCKET) {
				return 0;
			}
			if (!OpenClipboard(hwnd)) {
				return 0;
			}
			if (NULL == (hClipText = GetClipboardData(CF_TEXT))) {
				CloseClipboard();
				return 0;
			}
			if (NULL == (pszPasteText = (char *)malloc(GlobalSize(hClipText)))) {
				CloseClipboard();
				return 0;
			}
			pClipText = GlobalLock(hClipText);
			strcpy(pszPasteText, pClipText);
			GlobalUnlock(hClipText);
			CloseClipboard();
			if (pszPasteBuffer != NULL) {
				free(pszPasteBuffer);
			}
			pszPasteBuffer = ConvertCodeSet(pszPasteText);
			if (pszPasteBuffer == NULL) {
				free(pszPasteText);
				break;
			}
			nPasteLength   = strlen(pszPasteBuffer);
			nPastePosition = 0;
			do {
				rc = send(Soc, &pszPasteBuffer[nPastePosition], nPasteLength, 0);
				if (rc == SOCKET_ERROR) {
					if (WSAGetLastError() == WSAEWOULDBLOCK) {
						return 0;
					}
					break;
				} else {
					nPasteLength   -= rc;
					nPastePosition += rc;
				}
			} while (nPasteLength > 0);
			free(pszPasteBuffer);
			pszPasteBuffer = NULL;
			break;
		case IDM_CLEARBUFFER:
			pLineBuf = pMonTopLine->pPrevLine;
			while (pLineBuf != NULL) {
				pPrevLine = pLineBuf->pPrevLine;
				FreeLineBuf(pLineBuf);
				pLineBuf = pPrevLine;
			}
			pMonTopLine->pPrevLine = NULL;
			pTopLine = pViewTopLine = pMonTopLine;
			iViewTopLine = 0;
			xSelectFrom = ySelectFrom = xSelectTo = ySelectTo = 0;
			InvalidateRect(hwnd, NULL, TRUE);
			vsi.fMask = SIF_POS | SIF_RANGE | SIF_DISABLENOSCROLL;
			vsi.nMax  = nROWS - 1;
			vsi.nPos  = vsi.nMax - max((int)vsi.nPage - 1, 0);
			SetScrollInfo(hwnd, SB_VERT, &vsi, TRUE);
			SetCaret();
			break;
		case IDM_CHOOSEFONT:
			if (OpenChooseFontDlg(hwnd, &logfont)) {
				DeleteObject(hFont);
				hFont = CreateFontIndirect(&logfont);
				iSpace = GetFontCharSize(hwnd, hFont, &cxChar, &cyChar);
				ChangeFontVT(hwnd, hFont);
				SetWindowPos(hwnd, NULL, 0, 0,
							 cxChar * nCOLS + GetSystemMetrics(SM_CXSIZEFRAME) * 2 +
							 GetSystemMetrics(SM_CXEDGE) * 2 +
							 GetSystemMetrics(SM_CXVSCROLL),
							 cyChar * nROWS + GetSystemMetrics(SM_CYSIZEFRAME) * 2 +
							 GetSystemMetrics(SM_CYEDGE) * 2 +
							 GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU),
							 SWP_NOMOVE | SWP_NOZORDER);
				iCaretSize = 0;// Lbg蒼
				SetCaret();
				InvalidateRect(hwnd, NULL, TRUE);
				RegWriteFont(&logfont);
			} else {
				// ꂽlogfont̏ɖ߂
				GetObject(hFont, sizeof(LOGFONT), (PVOID)&logfont);
			}
			break;
		case IDM_RESIZE:
			if (DialogBox(hInst, MAKEINTRESOURCE(IDD_TERMSIZE), hwnd, TermSizeDlgProc)) {
				SetWindowPos(hwnd, NULL, 0, 0,
							 cxChar * nCOLS + GetSystemMetrics(SM_CXSIZEFRAME) * 2 +
							 GetSystemMetrics(SM_CXEDGE) * 2 +
							 GetSystemMetrics(SM_CXVSCROLL),
							 cyChar * nROWS + GetSystemMetrics(SM_CYSIZEFRAME) * 2 +
							 GetSystemMetrics(SM_CYEDGE) * 2 +
							 GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU),
							 SWP_NOMOVE | SWP_NOZORDER);
				vsi.fMask  = SIF_PAGE | SIF_RANGE | SIF_DISABLENOSCROLL;
				vsi.nMax   = nLineCount - 1;
				vsi.nPage  = nROWS;
				SetScrollInfo(hwnd, SB_VERT, &vsi, TRUE);
				SetScrollRangeMon(0, nROWS - 1);
				InvalidateRect(hwnd, NULL, TRUE);
				x = x >= nCOLS ? nCOLS - 1 : x;
				y = y >= nROWS ? nROWS - 1 : y;
				SetCaret();
				if (Soc != INVALID_SOCKET) {
					SendNAWS();
				}
				RegWriteTermSize(nCOLS, nROWS, nBackLogMax);
			}
			break;
		case IDM_DEBUGTRACE:
			bDebugTrace = !bDebugTrace;
			break;
		case IDM_ABOUT:
			DialogBox(hInst, MAKEINTRESOURCE(IDD_ABOUT), hwnd, AboutDlgProc);
			break;
		case IDM_MYCOLOR:
			if (ChangeMyColor(hInst, hwnd)) {
				InvalidateRect(hwnd, NULL, TRUE);
			}
			break;
		case IDM_SELECTALL:
			xSelectFrom = 0;
			ySelectFrom = 0;
			xSelectTo   = nCOLS;
			ySelectTo   = nLineCount - 1;
			InvalidateRect(hwnd, NULL, TRUE);
			break;
		case IDM_EUC:
		case IDM_SJIS:
		case IDM_UTF8:
			iCodeSet = LOWORD(wParam) - IDM_EUC;
			break;
		}
		return 0;
	case WM_CREATE:
		// tHg̏ݒ
		if (!RegReadFont(&logfont)) {
			GetObject(GetStockObject(SYSTEM_FIXED_FONT), sizeof(LOGFONT), (PVOID)&logfont);
		}
		hFont = CreateFontIndirect(&logfont);
		iSpace = GetFontCharSize(hwnd, hFont, &cxChar, &cyChar);
		ChangeFontVT(hwnd, hFont);
		// f[^\̏ݒ
		for (i = 0; i < nROWS; i++) {
			pLineBuf = CreateNewLine(NULL, NULL);
			if (pLineBuf == NULL) {
				return -1;
			}
			if (pTopLine == NULL) {
				pTopLine = pLineBuf;
				pEndLine = pLineBuf;
			} else {
				pLineBuf->pPrevLine = pEndLine;
				pEndLine->pNextLine = pLineBuf;
				pEndLine = pLineBuf;
			}
		}
		pViewTopLine = pTopLine;
		pMonTopLine = pTopLine;
		// XN[o[(c)̏ݒ
		memset(&vsi, 0, sizeof(vsi));
		vsi.cbSize = sizeof(SCROLLINFO);
		vsi.fMask  = SIF_PAGE | SIF_POS | SIF_RANGE | SIF_DISABLENOSCROLL;
		vsi.nMin   = 0;
		vsi.nMax   = nROWS - 1;
		vsi.nPage  = nROWS;
		vsi.nPos   = vsi.nMax - max((int)vsi.nPage - 1, 0);
		SetScrollInfo(hwnd, SB_VERT, &vsi, TRUE);
		// LvV̏ݒ
		LoadString(hInst, IDS_DISCONNECTED, szString, sizeof(szString));
		FormatCaption(hwnd, szAppName, szString);
		return 0;
	case WM_CLOSE:
		if (Soc != INVALID_SOCKET) {
			LoadString(hInst, IDS_QUERYDISC, szString, sizeof(szString));
			if (IDNO == MessageBox(hwnd, szString, szAppName, MB_ICONEXCLAMATION | MB_YESNO)) {
				return 0;
			}
		}
		DestroyWindow(hwnd);
		return 0;
	case WM_QUERYENDSESSION:
		if (Soc != INVALID_SOCKET) {
			LoadString(hInst, IDS_QUERYDISC, szString, sizeof(szString));
			if (IDNO == MessageBox(hwnd, szString, szAppName, MB_ICONEXCLAMATION | MB_YESNO)) {
				return FALSE;
			}
		}
		return TRUE;
	case WM_ENDSESSION:
		if (wParam && Soc != INVALID_SOCKET) {
			shutdown(Soc, SD_BOTH);
			closesocket(Soc);
		}
		return 0;
	case WM_DESTROY:
		if (Soc != INVALID_SOCKET) {
			shutdown(Soc, SD_BOTH);
			closesocket(Soc);
		}
		DeleteObject(hFont);
		TermMyColor();
#ifdef _DEBUG
		pLineBuf = pEndLine;
		while (pLineBuf != NULL) {
			pPrevLine = pLineBuf->pPrevLine;
			FreeLineBuf(pLineBuf);
			pLineBuf = pPrevLine;
		}
#endif /* _DEBUG */
		PostQuitMessage(0);
		return 0;
	case WM_PAINT:
		hdc = BeginPaint(hwnd, &ps);
		iPaintBeg = max(0, ps.rcPaint.top / cyChar);
		iPaintEnd = min(nROWS - 1, ps.rcPaint.bottom / cyChar);
		SetSelArea(xSelectFrom, ySelectFrom, xSelectTo, ySelectTo, &rect);
		hOldFont = SelectObject(hdc, hFont);
		pLineBuf = pViewTopLine;
		for (i = 0; i < nROWS; i++) {
			if (i >= iPaintBeg && i <= iPaintEnd) {
				if (iViewTopLine + i == rect.top) {
					TextOutVT(hdc, 0, i * cyChar, pLineBuf->pszLine, rect.left,
							  hFont, cxChar, FALSE, pLineBuf->pchAttr);
					if (rect.top != rect.bottom) {
						TextOutVT(hdc, rect.left * cxChar, i * cyChar,
								  &pLineBuf->pszLine[rect.left], pLineBuf->iLen - rect.left,
								  hFont, cxChar, TRUE, &pLineBuf->pchAttr[rect.left]);
					} else {
						TextOutVT(hdc, rect.left * cxChar, i * cyChar,
								  &pLineBuf->pszLine[rect.left], rect.right - rect.left,
								  hFont, cxChar, TRUE, &pLineBuf->pchAttr[rect.left]);
						TextOutVT(hdc, rect.right * cxChar, i * cyChar,
								  &pLineBuf->pszLine[rect.right], pLineBuf->iLen - rect.right,
								  hFont, cxChar, FALSE, &pLineBuf->pchAttr[rect.right]);
					}
				} else if (iViewTopLine + i > rect.top && iViewTopLine + i < rect.bottom) {
					TextOutVT(hdc, 0, i * cyChar, pLineBuf->pszLine, pLineBuf->iLen,
							  hFont, cxChar, TRUE, pLineBuf->pchAttr);
				} else if (iViewTopLine + i == rect.bottom) {
					TextOutVT(hdc, 0, i * cyChar, pLineBuf->pszLine, rect.right,
							  hFont, cxChar, TRUE, pLineBuf->pchAttr);
					TextOutVT(hdc, rect.right * cxChar, i * cyChar,
							  &pLineBuf->pszLine[rect.right], pLineBuf->iLen - rect.right,
							  hFont, cxChar, FALSE, &pLineBuf->pchAttr[rect.right]);
				} else {
					TextOutVT(hdc, 0, i * cyChar, pLineBuf->pszLine, pLineBuf->iLen,
							  hFont, cxChar, FALSE, pLineBuf->pchAttr);
				}
			}
			pLineBuf = pLineBuf->pNextLine;
		}
		SelectObject(hdc, hOldFont);
		{
			HPEN hPen, hOldPen;

			hPen = CreatePen(PS_SOLID, 0, GetMyColor(COLOR_WINDOWTEXT));
			hOldPen = SelectObject(hdc, hPen);
			MoveToEx(hdc, 0, cyChar * nROWS, NULL);
			LineTo(hdc, cxChar * nCOLS, cyChar * nROWS);
			LineTo(hdc, cxChar * nCOLS, -1);
			SelectObject(hdc, hOldPen);
			DeleteObject(hPen);
		}
		EndPaint(hwnd, &ps);
		return 0;
	case WM_SETFOCUS:
		// create and show the caret
		CreateCaret(hwnd, NULL, cxChar * iCaretSize, cyChar);
		SetCaret();
		return 0;
	case WM_KILLFOCUS:
		// hide and destroy the caret
		if (!bHideCaret) {
			HideCaret(hwnd);
			bHideCaret = TRUE;
		}
		DestroyCaret();
		return 0;
	case WM_IME_STARTCOMPOSITION:
		if (hIMC = ImmGetContext(hwnd)) {
			cf.dwStyle = CFS_POINT;
			cf.ptCurrentPos.x = max(0, min(x * cxChar, nCOLS * cxChar));
			cf.ptCurrentPos.y = max(0, min(y * cyChar, nROWS * cyChar));
			ImmSetCompositionWindow(hIMC, &cf);
			ImmSetCompositionFont(hIMC, &logfont);	// tHg̕ύX
			ImmReleaseContext(hwnd, hIMC);
		}
		break;
	case WM_VSCROLL:
		switch (LOWORD(wParam)) {
		case SB_TOP:
			iScrollInc = -vsi.nPos;
			break;
		case SB_BOTTOM:
			iScrollInc = vsi.nMax - max(vsi.nPage - 1, 0);
			break;
		case SB_LINEUP:
			iScrollInc = -1;
			break;
		case SB_LINEDOWN:
			iScrollInc = 1;
			break;
		case SB_PAGEUP:
			iScrollInc = -(int)vsi.nPage;
			break;
		case SB_PAGEDOWN:
			iScrollInc = vsi.nPage;
			break;
		case SB_THUMBTRACK:
			{
				SCROLLINFO si;

				si.cbSize = sizeof(SCROLLINFO);
				si.fMask = SIF_TRACKPOS;
				GetScrollInfo(hwnd, SB_VERT, &si);
				iScrollInc = si.nTrackPos - vsi.nPos;
			}
			bThumbTrack = TRUE;
			break;
		case SB_THUMBPOSITION:
			bThumbTrack = FALSE;/* (X[) */
		default:
			iScrollInc = 0;
			break;
		}
		iScrollInc = max(-vsi.nPos,
						 min(iScrollInc, vsi.nMax - max((int)vsi.nPage - 1, 0) - vsi.nPos));
		if (iScrollInc != 0) {
			for (i = 0; i > iScrollInc; i--) {
				pViewTopLine = pViewTopLine->pPrevLine;
				iViewTopLine--;
			}
			for (i = 0; i < iScrollInc; i++) {
				pViewTopLine = pViewTopLine->pNextLine;
				iViewTopLine++;
			}
			SetRect(&rect, 0, 0, cxChar * nCOLS, cyChar * nROWS);
			ScrollWindowEx(hwnd, 0, -cyChar * iScrollInc, &rect, &rect, NULL, NULL,
						   SW_ERASE | SW_INVALIDATE);
			vsi.fMask = SIF_POS;
			vsi.nPos += iScrollInc;
			SetScrollInfo(hwnd, SB_VERT, &vsi, TRUE);
			SetCaret();
			UpdateWindow(hwnd);
		}
		return 0;
	case WM_LBUTTONDOWN:
		// ȑOIĂ̈Ҕ
		SetSelArea(xSelectFrom, ySelectFrom, xSelectTo, ySelectTo, &rect);

		xSelectFrom = min(max(0, (short int)LOWORD(lParam) / cxChar), nCOLS);
		ySelectFrom = min(max(0, (short int)HIWORD(lParam) / cyChar), nROWS - 1);
		// 2byte̊ԂŐ؂Ȃ悤ɒ߂
		pLineBuf = pViewTopLine;
		for (i = 0; i < ySelectFrom; i++) {
			pLineBuf = pLineBuf->pNextLine;
		}
		if (pLineBuf->iLen < xSelectFrom) {
			xSelectFrom = pLineBuf->iLen;
		}
		if (IsDBCSSecondByte(pLineBuf->pszLine, xSelectFrom)) {
			xSelectFrom--;
		}
		ySelectFrom += iViewTopLine;// XN[Z
		xSelectTo = xSelectFrom;
		ySelectTo = ySelectFrom;
		SetCapture(hwnd);
		SetTimer(hwnd, 1, 100, NULL);
		bLeftButtonDown = TRUE;
		// ȑOIĂ̈ĕ`悷(I)
		InvalidateSelArea(hwnd, &rect, -(iViewTopLine * cyChar));
		return 0;
	case WM_LBUTTONUP:
		if (bLeftButtonDown) {
			ReleaseCapture();
			KillTimer(hwnd, 1);
			bLeftButtonDown = FALSE;
		}
		bWordSelect = FALSE;
		return 0;
	case WM_LBUTTONDBLCLK:// DOWNUPDBLCLKUP
		pLineBuf = pViewTopLine;
		for (i = 0; i < ySelectFrom - iViewTopLine; i++) {
			pLineBuf = pLineBuf->pNextLine;
		}
		GetWordPos(pLineBuf->pszLine, xSelectFrom, &iWordFrom, &iWordTo);
		xSelectFrom = iWordFrom;
		xSelectTo   = iWordTo;
		// IꂽPĕ`悷
		SetSelArea(xSelectFrom, ySelectFrom, xSelectTo, ySelectTo, &rect);
		InvalidateSelArea(hwnd, &rect, -(iViewTopLine * cyChar));
		bWordSelect = TRUE;
		SetCapture(hwnd);
		SetTimer(hwnd, 1, 100, NULL);
		bLeftButtonDown = TRUE;
		return 0;
	case WM_RBUTTONDOWN:
		// Rs[  I  \t
		SendMessage(hwnd, WM_COMMAND, IDM_COPY, 0L);
		SetSelArea(xSelectFrom, ySelectFrom, xSelectTo, ySelectTo, &rect);
		xSelectFrom = ySelectFrom = xSelectTo = ySelectTo = 0;
		InvalidateSelArea(hwnd, &rect, -(iViewTopLine * cyChar));
		SendMessage(hwnd, WM_COMMAND, IDM_PASTE, 0L);
		return 0;
	case WM_MOUSEMOVE:
		if (bLeftButtonDown) {
			RECT rect1, rect2;

			// ȑOIĂ̈Ҕ
			SetSelArea(xSelectFrom, ySelectFrom, xSelectTo, ySelectTo, &rect1);

			xSelectTo = min(max(0, (short int)LOWORD(lParam) / cxChar), nCOLS);
			ySelectTo = min(max(0, (short int)HIWORD(lParam) / cyChar), nROWS - 1);
			// 2byte̊ԂŐ؂Ȃ悤ɒ߂
			pLineBuf = pViewTopLine;
			for (i = 0; i < ySelectTo; i++) {
				pLineBuf = pLineBuf->pNextLine;
			}
			if (pLineBuf->iLen < xSelectTo) {
				xSelectTo = pLineBuf->iLen;
			}
			if (IsDBCSSecondByte(pLineBuf->pszLine, xSelectTo)) {
				xSelectTo--;
			}
			ySelectTo += iViewTopLine;// XN[Z

			if (bWordSelect) {// _uNbNɂPĨhbO
				if (ySelectFrom < ySelectTo) {
					xSelectFrom = iWordFrom;
				} else if (ySelectFrom > ySelectTo) {
					xSelectFrom = iWordTo;
				} else {
					if (xSelectTo < iWordFrom) {
						xSelectFrom = iWordTo;
					} else {
						xSelectFrom = iWordFrom;
						xSelectTo   = max(xSelectTo, iWordTo);
					}
				}
			}
			// ύXꂽÏ̍ĕ`悷
			SetSelArea(xSelectFrom, ySelectFrom, xSelectTo, ySelectTo, &rect2);
			InvalidateSelChanged(hwnd, &rect1, &rect2, -(iViewTopLine * cyChar));
		}
		break;
	case WM_MOUSEWHEEL:
		{
			WORD       fwKeys = LOWORD(wParam);
			short      zDelta = (short)HIWORD(wParam);
			UINT       vParam;
			UINT       uBar;
			int        nScrollCode;
			UINT       uScrollCnt;
			SCROLLINFO si;

			if (fwKeys & MK_CONTROL || fwKeys & MK_SHIFT) {
				return 0;
			}
			memset(&si, 0, sizeof(si));
			si.cbSize = sizeof(si);
			si.fMask  = SIF_PAGE | SIF_RANGE;
			GetScrollInfo(hwnd, SB_VERT, &si);// SigTerm ͉XN[Ȃ...
			uBar = si.nMax < (int)si.nPage ? WM_HSCROLL : WM_VSCROLL;
			SystemParametersInfo(SPI_GETWHEELSCROLLLINES, 0, &vParam, 0);
			if (vParam == WHEEL_PAGESCROLL) {
				SendMessage(hwnd, uBar, zDelta < 0 ? SB_PAGEDOWN : SB_PAGEUP, 0);
			} else {
				nScrollCode = zDelta < 0 ? SB_LINEDOWN : SB_LINEUP;
				vParam = abs(zDelta) / WHEEL_DELTA * vParam;
				for (uScrollCnt = 0; uScrollCnt < vParam; uScrollCnt++) {
					SendMessage(hwnd, uBar, nScrollCode, 0);
				}
			}
		}
		return 0;
	case WM_TIMER:
		GetCursorPos(&point);
		ScreenToClient(hwnd, &point);
		GetClientRect(hwnd, &rect);
		if (point.y > rect.bottom && vsi.nPos < vsi.nMax - (int)vsi.nPage + 1) {
			SendMessage(hwnd, WM_VSCROLL, SB_LINEDOWN, 0L);
			SendMessage(hwnd, WM_MOUSEMOVE, MK_LBUTTON, MAKELONG(point.x, point.y));
		} else if (point.y < 0 && vsi.nPos > 0) {
			SendMessage(hwnd, WM_VSCROLL, SB_LINEUP, 0L);
			SendMessage(hwnd, WM_VSCROLL, SB_LINEUP, 0L);// ͒x̂2
			SendMessage(hwnd, WM_MOUSEMOVE, MK_LBUTTON, MAKELONG(point.x, point.y));
		}
		return 0;
	case WM_INITMENUPOPUP:
		if ((HMENU)wParam == GetSubMenu(GetSubMenu(GetMenu(hwnd), 0), 3)) {
			CheckMenuRadioItem((HMENU)wParam, IDM_EUC, IDM_UTF8, IDM_EUC + iCodeSet,
							   MF_BYCOMMAND);
			return 0;
		}
		switch (LOWORD(lParam)) {
		case 0:// [RlNV]j[
			if (Soc == INVALID_SOCKET) {
				EnableMenuItem((HMENU)wParam, IDM_CONNECT,    MF_BYCOMMAND | MF_ENABLED);
				EnableMenuItem((HMENU)wParam, IDM_DISCONNECT, MF_BYCOMMAND | MF_GRAYED);
			} else {
				EnableMenuItem((HMENU)wParam, IDM_CONNECT,    MF_BYCOMMAND | MF_GRAYED);
				EnableMenuItem((HMENU)wParam, IDM_DISCONNECT, MF_BYCOMMAND | MF_ENABLED);
			}
			break;
		case 1:// [ҏW]j[
			if (xSelectFrom != xSelectTo || ySelectFrom != ySelectTo) {
				EnableMenuItem((HMENU)wParam, IDM_COPY, MF_BYCOMMAND | MF_ENABLED);
			} else {
				EnableMenuItem((HMENU)wParam, IDM_COPY, MF_BYCOMMAND | MF_GRAYED);
			}
			if (Soc != INVALID_SOCKET && IsClipboardFormatAvailable(CF_TEXT)) {
				EnableMenuItem((HMENU)wParam, IDM_PASTE, MF_BYCOMMAND | MF_ENABLED);
			} else {
				EnableMenuItem((HMENU)wParam, IDM_PASTE, MF_BYCOMMAND | MF_GRAYED);
			}
			break;
		case 3:// [wv]j[
			if (bDebugTrace) {
				CheckMenuItem((HMENU)wParam, IDM_DEBUGTRACE, MF_BYCOMMAND | MF_CHECKED);
			} else {
				CheckMenuItem((HMENU)wParam, IDM_DEBUGTRACE, MF_BYCOMMAND | MF_UNCHECKED);
			}
			break;
		}
		return 0;
	case WM_SETTINGCHANGE:
		SetWindowPos(hwnd, NULL, 0, 0,
					 cxChar * nCOLS + GetSystemMetrics(SM_CXSIZEFRAME) * 2 +
					 GetSystemMetrics(SM_CXEDGE) * 2 +
					 GetSystemMetrics(SM_CXVSCROLL),
					 cyChar * nROWS + GetSystemMetrics(SM_CYSIZEFRAME) * 2 +
					 GetSystemMetrics(SM_CYEDGE) * 2 +
					 GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU),
					 SWP_NOMOVE | SWP_NOZORDER);
		return 0;
	}
	return DefWindowProc(hwnd, message, wParam, lParam);
}

/* Ԋu𒲐TextOut */
BOOL AdjustTextOut(HDC hdc, int nXStart, int nYStart, LPCTSTR lpString, int cbString,
				   int iExtra1, int iExtra2)
{
	int  i;					// JE^
	char *pStartPos;		// ZeXJnʒuւ̃|C^
	BYTE *pCurrentChar;		// ݕւ̃|C^
	int  iSentenceLen;		// ZeX̒
	BOOL bDbcsString;		// 2BytetO

	if (iSpace == 0 && iExtra1 == iExtra2) {
		SetTextCharacterExtra(hdc, - iExtra1);
		return TextOut(hdc, nXStart, nYStart, lpString, cbString);
	}
	iSentenceLen = 0;
	pStartPos    = (char *)lpString;
	pCurrentChar = pStartPos;
	for (i = 0; i < cbString; i++) {
		if (IsDBCSLeadByte(*pCurrentChar)) {
			if (iSentenceLen > 0 && !bDbcsString) {
				SetTextCharacterExtra(hdc, - iExtra1);
				TextOut(hdc, nXStart, nYStart, pStartPos, iSentenceLen);
				pStartPos = pCurrentChar;
				nXStart = nXStart + iSentenceLen * cxChar;
				iSentenceLen = 0;
			}
			bDbcsString = TRUE;
			iSentenceLen += 2;
			pCurrentChar += 2;
			i++;
		} else {
			if (iSentenceLen > 0 && bDbcsString) {
				SetTextCharacterExtra(hdc, iSpace - iExtra2);
				TextOut(hdc, nXStart, nYStart, pStartPos, iSentenceLen);
				pStartPos = pCurrentChar;
				nXStart = nXStart + iSentenceLen * cxChar;
				iSentenceLen = 0;
			}
			bDbcsString = FALSE;
			iSentenceLen++;
			pCurrentChar++;
		}
	}
	if (iSentenceLen > 0) {
		SetTextCharacterExtra(hdc, bDbcsString ? iSpace - iExtra2 : - iExtra1);
		TextOut(hdc, nXStart, nYStart, pStartPos, iSentenceLen);
	}
	return TRUE;
}
