/*-------------------------------------------------------
   MYCOLOR.C -- Telnet Client, Terminal Emulator Program
             zF
   Copyright (c) 2003
        Hiroto Shigeta. All rights reserved.
  -------------------------------------------------------*/
#include <windows.h>
#include "resource.h"

static BOOL        bDefColor = TRUE;
static COLORREF    crWndBack = RGB(255, 255, 255);	// eLXg̔wiF
static COLORREF    crWndText = RGB(0, 0, 0);		// eLXg̐F
static COLORREF    crHltBack = RGB(0, 0, 0);		// IeLXg̔wiF
static COLORREF    crHltText = RGB(255, 255, 255);	// IeLXg̐F
static COLORREF    crCustColors[16];				// ŝݗLȎF
static CHOOSECOLOR cc;
static HBRUSH      hbrBackground = NULL;			// wiF̃uV
static char        szRegSubKey[64];					// WXgTuL[

static BOOL RegReadMyColor(void)
{
	HKEY  hkResult;
	DWORD cbData;

	if (RegOpenKeyEx(HKEY_CURRENT_USER, szRegSubKey, 0,
					 KEY_READ, &hkResult) != ERROR_SUCCESS) {
		return FALSE;
	}
	cbData = sizeof(BOOL);
	RegQueryValueEx(hkResult, "color default", NULL, NULL, (LPBYTE)&bDefColor, &cbData);
	cbData = sizeof(COLORREF);
	RegQueryValueEx(hkResult, "color window", NULL, NULL, (LPBYTE)&crWndBack, &cbData);
	cbData = sizeof(COLORREF);
	RegQueryValueEx(hkResult, "color window text", NULL, NULL, (LPBYTE)&crWndText, &cbData);
	cbData = sizeof(COLORREF);
	RegQueryValueEx(hkResult, "color highlight", NULL, NULL, (LPBYTE)&crHltBack, &cbData);
	cbData = sizeof(COLORREF);
	RegQueryValueEx(hkResult, "color highlight text", NULL, NULL, (LPBYTE)&crHltText, &cbData);
	RegCloseKey(hkResult);
	return TRUE;
}

static BOOL RegWriteMyColor(void)
{
	HKEY  hkResult;
	DWORD dwDisposition;

	if (RegCreateKeyEx(HKEY_CURRENT_USER, szRegSubKey, 0, NULL,
					   REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL,
					   &hkResult, &dwDisposition) != ERROR_SUCCESS) {
		return FALSE;
	}
	RegSetValueEx(hkResult, "color default", 0, REG_DWORD, (LPBYTE)&bDefColor,
				  sizeof(BOOL));
	RegSetValueEx(hkResult, "color window", 0, REG_BINARY, (LPBYTE)&crWndBack,
				  sizeof(COLORREF));
	RegSetValueEx(hkResult, "color window text", 0, REG_BINARY, (LPBYTE)&crWndText,
				  sizeof(COLORREF));
	RegSetValueEx(hkResult, "color highlight", 0, REG_BINARY, (LPBYTE)&crHltBack,
				  sizeof(COLORREF));
	RegSetValueEx(hkResult, "color highlight text", 0, REG_BINARY, (LPBYTE)&crHltText,
				  sizeof(COLORREF));
	RegCloseKey(hkResult);
	return TRUE;
}

BOOL InitMyColor(LPCTSTR lpSubKey)
{
	strcpy(szRegSubKey, lpSubKey);
	return RegReadMyColor();
}

HBRUSH GetMyBkBrush(void)
{
	if (bDefColor) {
		return (HBRUSH)(COLOR_WINDOW + 1);
	} else {
		if (hbrBackground == NULL) {
			hbrBackground = CreateSolidBrush(crWndBack);
		}
		return hbrBackground;
	}
}

COLORREF GetMyColor(int nIndex)
{
	if (bDefColor) {
		return GetSysColor(nIndex);
	}
	switch (nIndex) {
	case COLOR_WINDOW:
		return crWndBack;
	case COLOR_WINDOWTEXT:
		return crWndText;
	case COLOR_HIGHLIGHT:
		return crHltBack;
	case COLOR_HIGHLIGHTTEXT:
		return crHltText;
	}
	return 0;
}

static BOOL OpenChooseColorDlg(HWND hwnd, COLORREF cr)
{
	cc.lStructSize    = sizeof(CHOOSECOLOR);
	cc.hwndOwner      = hwnd;
	cc.hInstance      = NULL;
	cc.rgbResult      = cr;
	cc.lpCustColors   = crCustColors;
	cc.Flags          = CC_RGBINIT;
	cc.lCustData      = 0L;
	cc.lpfnHook       = NULL;
	cc.lpTemplateName = NULL;

	return ChooseColor(&cc);
}

static BOOL CALLBACK ColorDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	static COLORREF crWorkWB, crWorkWT, crWorkHB, crWorkHT;
	static BOOL     bWorkDef;

	switch (uMsg) {
	case WM_MEASUREITEM:
		return TRUE;
	case WM_DRAWITEM:
		if (wParam == IDC_PREVIEW) {
			HBRUSH           hBrush;
			LPDRAWITEMSTRUCT lpdis  = (LPDRAWITEMSTRUCT)lParam;
			LOGFONT          logfont;
			HFONT            hFont, hOldFont;
			TEXTMETRIC       tm;

			hBrush = bWorkDef ? (HBRUSH)(COLOR_WINDOW + 1) : CreateSolidBrush(crWorkWB);
			FillRect(lpdis->hDC, &lpdis->rcItem, hBrush);
			if (!bWorkDef) {
				DeleteObject(hBrush);
			}
			GetObject(GetStockObject(SYSTEM_FIXED_FONT), sizeof(LOGFONT), (PVOID)&logfont);
			hFont = CreateFontIndirect(&logfont);
			hOldFont = SelectObject(lpdis->hDC, hFont);
			GetTextMetrics(lpdis->hDC, &tm);
			SetTextColor(lpdis->hDC, bWorkDef ? GetSysColor(COLOR_WINDOWTEXT) : crWorkWT);
			SetBkColor(lpdis->hDC, bWorkDef ? GetSysColor(COLOR_WINDOW) : crWorkWB);
			TextOut(lpdis->hDC, 0, 0, "venus% ls", 9);
			TextOut(lpdis->hDC, 0, tm.tmHeight, "mimi", 4);
			TextOut(lpdis->hDC, 0, tm.tmHeight * 2, "venus%", 6);
			SetTextColor(lpdis->hDC, bWorkDef ? GetSysColor(COLOR_HIGHLIGHTTEXT) : crWorkHT);
			SetBkColor(lpdis->hDC, bWorkDef ? GetSysColor(COLOR_HIGHLIGHT) : crWorkHB);
			TextOut(lpdis->hDC, tm.tmAveCharWidth * 8, tm.tmHeight, "tama", 4);
			SelectObject(lpdis->hDC, hOldFont);
			DeleteObject(hFont);
		}
		return TRUE;
	case WM_INITDIALOG:
		crWorkWB = crWndBack;
		crWorkWT = crWndText;
		crWorkHB = crHltBack;
		crWorkHT = crHltText;
		bWorkDef  = bDefColor;
		if (bWorkDef) {
			CheckDlgButton(hwndDlg, IDC_DEFCOLOR, TRUE);
			EnableWindow(GetDlgItem(hwndDlg, IDC_CRWNDTEXT), FALSE);
			EnableWindow(GetDlgItem(hwndDlg, IDC_CRWNDBACK), FALSE);
			EnableWindow(GetDlgItem(hwndDlg, IDC_CRHLTTEXT), FALSE);
			EnableWindow(GetDlgItem(hwndDlg, IDC_CRHLTBACK), FALSE);
		}
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case IDC_DEFCOLOR:
			bWorkDef = IsDlgButtonChecked(hwndDlg, IDC_DEFCOLOR);
			EnableWindow(GetDlgItem(hwndDlg, IDC_CRWNDTEXT), !bWorkDef);
			EnableWindow(GetDlgItem(hwndDlg, IDC_CRWNDBACK), !bWorkDef);
			EnableWindow(GetDlgItem(hwndDlg, IDC_CRHLTTEXT), !bWorkDef);
			EnableWindow(GetDlgItem(hwndDlg, IDC_CRHLTBACK), !bWorkDef);
			break;
		case IDC_CRWNDBACK:
			if (OpenChooseColorDlg(hwndDlg, crWorkWB)) {
				crWorkWB = cc.rgbResult;
			}
			break;
		case IDC_CRWNDTEXT:
			if (OpenChooseColorDlg(hwndDlg, crWorkWT)) {
				crWorkWT = cc.rgbResult;
			}
			break;
		case IDC_CRHLTBACK:
			if (OpenChooseColorDlg(hwndDlg, crWorkHB)) {
				crWorkHB = cc.rgbResult;
			}
			break;
		case IDC_CRHLTTEXT:
			if (OpenChooseColorDlg(hwndDlg, crWorkHT)) {
				crWorkHT = cc.rgbResult;
			}
			break;
		case IDOK:
			crWndBack = crWorkWB;
			crWndText = crWorkWT;
			crHltBack = crWorkHB;
			crHltText = crWorkHT;
			bDefColor = bWorkDef;
			EndDialog(hwndDlg, 1);
			return TRUE;
		case IDCANCEL:
			EndDialog(hwndDlg, 0);
			return TRUE;
		}
		InvalidateRect(GetDlgItem(hwndDlg, IDC_PREVIEW), NULL, TRUE);
		break;
	}
	return FALSE;
}

BOOL ChangeMyColor(HINSTANCE hInst, HWND hwnd)
{
	if (DialogBox(hInst, MAKEINTRESOURCE(IDD_COLORDLG), hwnd, ColorDialogProc)) {
		if (hbrBackground != NULL) {
			DeleteObject(hbrBackground);
			hbrBackground = NULL;
		}
		if (!bDefColor) {
			hbrBackground = CreateSolidBrush(crWndBack);
		}
		SetClassLong(hwnd, GCL_HBRBACKGROUND,
					 bDefColor ? (COLOR_WINDOW + 1) : (LONG)hbrBackground);
		RegWriteMyColor();
		return TRUE;
	}
	return FALSE;
}

void TermMyColor(void)
{
	if (hbrBackground != NULL) {
		DeleteObject(hbrBackground);
	}
}
