/*--------------------------------------------------------
   REGISTRY.C -- Telnet Client, Terminal Emulator Program
             WXg
   Copyright (c) 2002-2006
        Hiroto Shigeta. All rights reserved.
  --------------------------------------------------------*/
#include <windows.h>
#include "mycolor.h"

#define MRU_COUNT      10
//                      012345MRU_NUMBER_POS
#define MRU_VALUE_NAME "host 0"
#define MRU_NUMBER_POS 5

/* WXgTuL[ */
static char szRegSubKey[]    = "SOFTWARE\\hiroto\\SigTerm";
static char szRegSubKeyMRU[] = "SOFTWARE\\hiroto\\SigTerm\\most recently used";

BOOL RegReadFont(PLOGFONT pLogFont)
{
	HKEY  hkResult;
	DWORD cbData;
	BOOL  bRet = FALSE;

	if (RegOpenKeyEx(HKEY_CURRENT_USER, szRegSubKey, 0,
					 KEY_READ, &hkResult) != ERROR_SUCCESS) {
		return FALSE;
	}
	cbData = sizeof(LOGFONT);
	if (RegQueryValueEx(hkResult, "logfont", NULL, NULL, (LPBYTE)pLogFont,
						&cbData) == ERROR_SUCCESS) {
		bRet = TRUE;
	}
	RegCloseKey(hkResult);
	return bRet;
}

BOOL RegWriteFont(PLOGFONT pLogFont)
{
	HKEY  hkResult;
	DWORD dwDisposition;

	if (RegCreateKeyEx(HKEY_CURRENT_USER, szRegSubKey, 0, NULL,
					   REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL,
					   &hkResult, &dwDisposition) != ERROR_SUCCESS) {
		return FALSE;
	}
	RegSetValueEx(hkResult, "logfont", 0, REG_BINARY, (LPBYTE)pLogFont, sizeof(LOGFONT));
	RegCloseKey(hkResult);
	return TRUE;
}

BOOL RegReadTermSize(int *piCols, int *piRows, UINT *pnBackLogMax)
{
	HKEY  hkResult;
	DWORD cbData;

	if (RegOpenKeyEx(HKEY_CURRENT_USER, szRegSubKey, 0,
					 KEY_READ, &hkResult) != ERROR_SUCCESS) {
		InitMyColor(szRegSubKey);
		return FALSE;
	}
	cbData = sizeof(int);
	RegQueryValueEx(hkResult, "columns", NULL, NULL, (LPBYTE)piCols, &cbData);
	cbData = sizeof(int);
	RegQueryValueEx(hkResult, "rows", NULL, NULL, (LPBYTE)piRows, &cbData);
	cbData = sizeof(UINT);
	RegQueryValueEx(hkResult, "back log max", NULL, NULL, (LPBYTE)pnBackLogMax, &cbData);
	RegCloseKey(hkResult);
	InitMyColor(szRegSubKey);
	return TRUE;
}

BOOL RegWriteTermSize(int iCols, int iRows, UINT nBackLogMax)
{
	HKEY  hkResult;
	DWORD dwDisposition;

	if (RegCreateKeyEx(HKEY_CURRENT_USER, szRegSubKey, 0, NULL,
					   REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL,
					   &hkResult, &dwDisposition) != ERROR_SUCCESS) {
		return FALSE;
	}
	RegSetValueEx(hkResult, "columns", 0, REG_DWORD, (LPBYTE)&iCols, sizeof(int));
	RegSetValueEx(hkResult, "rows", 0, REG_DWORD, (LPBYTE)&iRows, sizeof(int));
	RegSetValueEx(hkResult, "back log max", 0, REG_DWORD, (LPBYTE)&nBackLogMax, sizeof(UINT));
	RegCloseKey(hkResult);
	return TRUE;
}

BOOL RegReadAutoClose(BOOL *pbAutoClose)
{
	HKEY  hkResult;
	DWORD cbData;

	if (RegOpenKeyEx(HKEY_CURRENT_USER, szRegSubKey, 0,
					 KEY_READ, &hkResult) != ERROR_SUCCESS) {
		return FALSE;
	}
	cbData = sizeof(BOOL);
	RegQueryValueEx(hkResult, "auto close", NULL, NULL, (LPBYTE)pbAutoClose, &cbData);
	RegCloseKey(hkResult);
	return TRUE;
}

BOOL RegWriteAutoClose(BOOL bAutoClose)
{
	HKEY  hkResult;
	DWORD dwDisposition;

	if (RegCreateKeyEx(HKEY_CURRENT_USER, szRegSubKey, 0, NULL,
					   REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL,
					   &hkResult, &dwDisposition) != ERROR_SUCCESS) {
		return FALSE;
	}
	RegSetValueEx(hkResult, "auto close", 0, REG_DWORD, (LPBYTE)&bAutoClose, sizeof(BOOL));
	RegCloseKey(hkResult);
	return TRUE;
}

BOOL InitMRU(HWND hCombo)
{
	HKEY  hkResult;
	int   i;
	char  szValueName[] = MRU_VALUE_NAME;
	char  szMRU[MAX_PATH];					// most recently used host list
	DWORD cbData;

	// WXgMRUǂݏo
	if (RegOpenKeyEx(HKEY_CURRENT_USER, szRegSubKeyMRU, 0,
					 KEY_READ, &hkResult) != ERROR_SUCCESS) {
		return FALSE;
	}
	for (i = 0; i < MRU_COUNT; i++) {
		szMRU[0] = '\0';
		szMRU[1] = '\0';
		cbData = MAX_PATH;// cbDataRegQueryValueEx()ōXV̂ŖZbg
		RegQueryValueEx(hkResult, szValueName, NULL, NULL, szMRU, &cbData);
		if (szMRU[0] == '\0') {
			break;
		}
		// zXgR{{bNXɒǉ
		SendMessage(hCombo, CB_ADDSTRING, 0, (LPARAM)szMRU);
		szValueName[MRU_NUMBER_POS]++;
	}
	RegCloseKey(hkResult);
	return TRUE;
}

int FindMRU(const char *pszHostName, int *piCodeSet, char *pszTermType, size_t cchTermType)
{
	HKEY  hkResult;
	int   i;
	char  szValueName[] = MRU_VALUE_NAME;
	char  szMRU[MAX_PATH];					// most recently used host list
	DWORD cbData;

	// WXgMRUǂݏo
	if (RegOpenKeyEx(HKEY_CURRENT_USER, szRegSubKeyMRU, 0,
					 KEY_READ, &hkResult) != ERROR_SUCCESS) {
		return -1;
	}
	for (i = 0; i < MRU_COUNT; i++) {
		szMRU[0] = '\0';
		szMRU[1] = '\0';
		cbData = MAX_PATH;// cbDataRegQueryValueEx()ōXV̂ŖZbg
		RegQueryValueEx(hkResult, szValueName, NULL, NULL, szMRU, &cbData);
		if (szMRU[0] == '\0') {
			break;
		}
		if (strcmp(pszHostName, szMRU) == 0) {
			RegCloseKey(hkResult);
			if (piCodeSet != NULL) {
				*piCodeSet = szMRU[strlen(szMRU) + 1] - '0';
			}
			if (pszTermType != NULL) {
				memset(pszTermType, 0, cchTermType);
				strncpy(pszTermType, szMRU + strlen(szMRU) + 3, cchTermType - 1);
			}
			return i;
		}
		szValueName[MRU_NUMBER_POS]++;
	}
	RegCloseKey(hkResult);
	return -1;
}

static size_t mszlen(const char *s)
{
	size_t count = 0;

	while (*s != '\0' || *(s + 1) != '\0') {
		count++;
		s++;
	}
	return count;
}

BOOL AddMRU(const char *pszHostName)
{
	int   i;
	HKEY  hkResult;
	DWORD dwDisposition;
	char  szValueName[] = MRU_VALUE_NAME;
	char  szMRU[MRU_COUNT][MAX_PATH];		// most recently used host list
	DWORD cbData;
	BOOL  bHostExists = FALSE;

	// Ǐϐ̏
	for (i = 0; i < MRU_COUNT; i++) {
		szMRU[i][0] = '\0';
		szMRU[i][1] = '\0';
	}
	// WXgMRUǂݏo
	if (RegCreateKeyEx(HKEY_CURRENT_USER, szRegSubKeyMRU, 0, NULL,
					   REG_OPTION_NON_VOLATILE, KEY_READ | KEY_WRITE, NULL,
					   &hkResult, &dwDisposition) != ERROR_SUCCESS) {
		return FALSE;
	}
	for (i = 0; i < MRU_COUNT; i++) {
		cbData = MAX_PATH;// cbDataRegQueryValueEx()ōXV̂ŖZbg
		RegQueryValueEx(hkResult, szValueName, NULL, NULL, szMRU[i], &cbData);
		szValueName[MRU_NUMBER_POS]++;
	}
	// ǉzXgMRUXgɊɑ݂ꍇ͍폜(Ԃl߂)
	for (i = 0; i < MRU_COUNT; i++) {
		if (strcmp(pszHostName, szMRU[i]) == 0) {
			for (; i > 0; i--) {
				memcpy(szMRU[i], szMRU[i - 1], MAX_PATH);
			}
			bHostExists = TRUE;
			break;
		}
	}
	// MRUXgVtg
	if (!bHostExists) {
		for (i = MRU_COUNT - 1; i > 0; i--) {
			memcpy(szMRU[i], szMRU[i - 1], MAX_PATH);
		}
	}
	// 擪ɒǉzXgZbg
	memcpy(szMRU[0], pszHostName, mszlen(pszHostName) + 2);
	// WXgɏ߂
	szValueName[MRU_NUMBER_POS] = '0';
	for (i = 0; i < MRU_COUNT; i++) {
		RegSetValueEx(hkResult, szValueName, 0, REG_EXPAND_SZ, szMRU[i], mszlen(szMRU[i]) + 2);
		szValueName[MRU_NUMBER_POS]++;
	}
	RegCloseKey(hkResult);
	return TRUE;
}

BOOL DelMRU(int iIndex)
{
	HKEY  hkResult;
	int   i;
	char  szValueName[] = MRU_VALUE_NAME;
	char  szMRU[MAX_PATH];					// most recently used host list
	DWORD cbData;

	// WXgMRUǂݏo
	if (RegOpenKeyEx(HKEY_CURRENT_USER, szRegSubKeyMRU, 0,
					 KEY_READ | KEY_WRITE, &hkResult) != ERROR_SUCCESS) {
		return FALSE;
	}
	// MRUzXg폜(Ԃl߂)
	szValueName[MRU_NUMBER_POS] += iIndex;
	for (i = iIndex + 1; i < MRU_COUNT; i++) {
		szValueName[MRU_NUMBER_POS]++;
		szMRU[0] = '\0';
		szMRU[1] = '\0';
		cbData = MAX_PATH;// cbDataRegQueryValueEx()ōXV̂ŖZbg
		RegQueryValueEx(hkResult, szValueName, NULL, NULL, szMRU, &cbData);
		szValueName[MRU_NUMBER_POS]--;
		RegSetValueEx(hkResult, szValueName, 0, REG_EXPAND_SZ, szMRU, mszlen(szMRU) + 2);
		szValueName[MRU_NUMBER_POS]++;
	}
	// ̃zXguNɂ
	szValueName[MRU_NUMBER_POS] = '0' + MRU_COUNT - 1;
	RegSetValueEx(hkResult, szValueName, 0, REG_EXPAND_SZ, "\0", 2);
	RegCloseKey(hkResult);
	return TRUE;
}
