/*------------------------------------------------------
   SOCKET.C -- Telnet Client, Terminal Emulator Program
             \Pbgڑ
   Copyright (c) 2002-2003
        Hiroto Shigeta. All rights reserved.
  ------------------------------------------------------*/
#include <winsock2.h>
#include <windows.h>
#include <stdio.h>
#include "socket.h"

/* \Pbg */
SOCKET Soc = INVALID_SOCKET;

/* fobOg[X */
BOOL bDebugTrace = FALSE;

/* \Pbgڑ */
int ConnectServer(HWND hwnd, const char *pszHost)
{
	PHOSTENT           pHostEntry;
	PSERVENT           pServEntry;
	struct sockaddr_in saServer;
	int                rc;
	int                iLastError;

	pHostEntry = gethostbyname(pszHost);
	if (pHostEntry == NULL) {
		unsigned long addr;

		addr = inet_addr(pszHost);
		pHostEntry = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);
		if (pHostEntry == NULL) {
			return WSAGetLastError();
		}
	}
	Soc = socket(AF_INET, SOCK_STREAM, 0);
	if (Soc == INVALID_SOCKET) {
		return WSAGetLastError();
	}
	rc = WSAAsyncSelect(Soc, hwnd, SM_EVENT,
						FD_CONNECT | FD_CLOSE | FD_READ | FD_WRITE);
	if (rc == SOCKET_ERROR) {
		iLastError = WSAGetLastError();
		closesocket(Soc);
		Soc = INVALID_SOCKET;
		return iLastError;
	}
	pServEntry = getservbyname("telnet", "tcp");
	if (pServEntry == NULL) {
		saServer.sin_port = htons(23);
	} else {
		saServer.sin_port = pServEntry->s_port;
	}
	saServer.sin_family = AF_INET;
	saServer.sin_addr = *((PIN_ADDR)*pHostEntry->h_addr_list);
	rc = connect(Soc, (PSOCKADDR)&saServer, sizeof(SOCKADDR));
	if (rc == SOCKET_ERROR) {
		iLastError = WSAGetLastError();
		if (iLastError != WSAEWOULDBLOCK) {
			closesocket(Soc);
			Soc = INVALID_SOCKET;
			return iLastError;
		}
	}
	return 0;
}

/* Mf[^fobOo */
int DebugRcvPrint(char *fmt, ...)
{
	if (bDebugTrace) {
		va_list args;
		FILE    *fp;
		char    fname[512];

		sprintf(fname, "%d.log", GetCurrentProcessId());
		fp = fopen(fname, "a");

		va_start(args, fmt);
		vfprintf(fp, fmt, args);
		va_end(args);

		fclose(fp);
	}
	return 0;
}

/* Mf[^fobOo */
int DebugSndPrint(char *fmt, ...)
{
	if (bDebugTrace) {
		va_list args;
		FILE    *fp;
		char    fname[512];

		sprintf(fname, "%d.log", GetCurrentProcessId());
		fp = fopen(fname, "a");

		va_start(args, fmt);
		vfprintf(fp, fmt, args);
		va_end(args);

		fclose(fp);
	}
	return 0;
}

/* 1oCgM */
int SendToServerOne(unsigned char c)
{
	send(Soc, &c, 1, 0);

	return 0;
}

/* 2oCgf[^M */
int SendToServerShort(short s)
{
	SendToServerOne((char)((s >> 8) & 0xFF));
	SendToServerOne((char)(s & 0xFF));

	return 0;
}

/* 1M(CRLFϊ) */
int SendToServerChar(char c)
{
	if (c == '\n') {
		SendToServerOne('\r');
	}
	SendToServerOne(c);

	return 0;
}

/* printfCNɑM */
int SendToServerPrint(char *fmt, ...)
{
	va_list args;
	char    buf[1024];
	int     len;
	int     pos;
	int     rc;

	va_start(args, fmt);
	vsprintf(buf, fmt, args);
	va_end(args);

	len = strlen(buf);
	pos = 0;
	do {
		rc = send(Soc, &buf[pos], len, 0);
		if (rc == SOCKET_ERROR) {
			break;
		} else {
			len -= rc;
			pos += rc;
		}
	} while (len > 0);

	return 0;
}
