/*------------------------------------------------------
   TELNET.C -- Telnet Client, Terminal Emulator Program
             TELNETvgR
   Copyright (c) 2002-2005
        Hiroto Shigeta. All rights reserved.
  ------------------------------------------------------*/
#include <winsock2.h>
#include <stdio.h>
#include "telnet.h"
#include "codeset.h"
#include "monitor.h"
#include "socket.h"
#include "terminal.h"

/* [JGR[ */
int Echo = 1;

/* :V */
int EnvNewOld = -1;

char cRecvState = 0;
char cCtrlState[32];

/* vg^Cv */
int IacDo(char c);
int IacWill(char c);
int IacDont(char c);
int IacWont(char c);
int IacSb(char c);
int SendTTYPE(void);
int SendXDISPLOC(void);
int SendNAWS(void);
int SendOLD_ENVIRON(void);
int SendNEW_ENVIRON(void);
int SendDefaultSb(char c);

/* Mf[^`FbN */
int RecvDataCheck(char c)
{
	static char chLeadByte = 0;

	if ((int)(cRecvState & 0xFF) == IAC) {
		if (cCtrlState[0] == 0) {
			switch ((int)(c & 0xFF)) {
			case DONT:
				DebugRcvPrint("[DONT]");
				cCtrlState[0] = c;
				break;
			case WONT:
				DebugRcvPrint("[WONT]");
				cCtrlState[0] = c;
				break;
			case DO:
				DebugRcvPrint("[DO]");
				cCtrlState[0] = c;
				break;
			case WILL:
				DebugRcvPrint("[WILL]");
				cCtrlState[0] = c;
				break;
			case SB:
				DebugRcvPrint("[SB]");
				cCtrlState[0] = c;
				break;
			case GA:
				DebugRcvPrint("[GA]");
				cRecvState = 0;
				break;
			case EL:
				DebugRcvPrint("[EL]");
				cRecvState = 0;
				break;
			case EC:
				DebugRcvPrint("[EC]");
				cRecvState = 0;
				break;
			case AYT:
				DebugRcvPrint("[AYT]");
				cRecvState = 0;
				break;
			case AO:
				DebugRcvPrint("[AO]");
				cRecvState = 0;
				break;
			case IP:
				DebugRcvPrint("[IP]");
				cRecvState = 0;
				break;
			case BREAK:
				DebugRcvPrint("[BREAK]");
				cRecvState = 0;
				break;
			case DM:
				DebugRcvPrint("[DM]");
				cRecvState = 0;
				break;
			case NOP:
				DebugRcvPrint("[NOP]");
				cRecvState = 0;
				break;
			case SE:
				DebugRcvPrint("[SE]");
				cRecvState = 0;
				break;
			case EOR:
				DebugRcvPrint("[EOR]");
				cRecvState = 0;
				break;
			case ABORT:
				DebugRcvPrint("[ABORT]");
				cRecvState = 0;
				break;
			case SUSP:
				DebugRcvPrint("[SUSP]");
				cRecvState = 0;
				break;
			case xEOF:
				DebugRcvPrint("[xEOF]");
				cRecvState = 0;
				break;
			default:
				DebugRcvPrint("[%03d]", (c & 0xFF));
				cRecvState = 0;
				break;
			}
			return 0;
		}
		/* R}h */
		switch ((int)(cCtrlState[0] & 0xFF)) {
		case DONT:
			IacDont(c);
			break;
		case WONT:
			IacWont(c);
			break;
		case DO:
			IacDo(c);
			break;
		case WILL:
			IacWill(c);
			break;
		case SB:
			IacSb(c);
			break;
		}
	} else if (cRecvState == '\x1B') {
		TerminalControl(c);
	} else if (chLeadByte != 0 && (c & 0xFF) >= (' ' & 0xFF)) {
		if (MonOutKanji(chLeadByte, c) == 0) {
			chLeadByte = 0;
		}
	} else {
		if ((int)(c & 0xFF) == IAC) {
			DebugRcvPrint("[IAC]");
			cRecvState = c;
			memset(&cCtrlState, 0, sizeof(cCtrlState));
			return 0;
		} else if (c == '\x1B') {
			DebugRcvPrint("[ESC]");
			cRecvState = c;
			memset(&cCtrlState, 0, sizeof(cCtrlState));
			return 0;
		} else if (IsHostCodeSetLeadByte(c)) {
			chLeadByte = c;
			return 0;
		}
		switch (c) {
		case '\r':
			DebugRcvPrint("[CR]");/* [^M] */
			MoveCaretLeftTop();
			break;
		case '\n':
			/* Lbgֈړ */
			DebugRcvPrint("[LF]\n");/* [^J] */
			MoveCaretDownOne();
			break;
		case '\b':
			/* obNXy[X */
			DebugRcvPrint("[^H]");
			MoveCaretLeftOne();
			break;
		case '\t':
			/* ^u */
			DebugRcvPrint("%c", c);/* [^I] */
			MoveCaretNextTab();
			break;
		default:
			/* ʂ */
			if (c == 0) {
				DebugRcvPrint("[NULL]");
				break;
			} else if (c == '\a') {
				DebugRcvPrint("[BEEP]");
				MessageBeep(MB_OK);
				break;
			} else if (IsShiftCode(&c)) {/* LinuxΉ+JIS7Ή */
				DebugRcvPrint("[%03d]", (c & 0xFF));
				break;
			}
			MonOutOne(c);
			DebugRcvPrint("%c", c);
			break;
		}
	}
	return 0;
}

/* DOR}h */
int IacDo(char c)
{
	switch ((int)(c & 0xFF)) {
	case TELOPT_ECHO:	/* Echo */
		DebugRcvPrint("[TELOPT_ECHO]\n");
		SendToServerOne(IAC);
		SendToServerOne(WILL);
		SendToServerOne(TELOPT_ECHO);
		DebugSndPrint("<IAC><WILL><TELOPT_ECHO>\n");
		Echo = 1;
		break;
	case TELOPT_TTYPE:	/* Terminal Type */
		DebugRcvPrint("[TELOPT_TTYPE]\n");
		SendToServerOne(IAC);
		SendToServerOne(WILL);
		SendToServerOne(TELOPT_TTYPE);
		DebugSndPrint("<IAC><WILL><TELOPT_TTYPE>\n");
		break;
	case TELOPT_NAWS:	/* Window Size */
		DebugRcvPrint("[TELOPT_NAWS]\n");
		SendToServerOne(IAC);
		SendToServerOne(WILL);
		SendToServerOne(TELOPT_NAWS);
		DebugSndPrint("<IAC><WILL><TELOPT_NAWS>\n");
		/* ʃTCYM */
		SendNAWS();
		break;
	case TELOPT_OLD_ENVIRON:	/* Old - Environment Variables */
		DebugRcvPrint("[TELOPT_OLD_ENVIRON]\n");
		if (EnvNewOld == -1) {
			SendToServerOne(IAC);
			SendToServerOne(WILL);
			SendToServerOne(TELOPT_OLD_ENVIRON);
			DebugSndPrint("<IAC><WILL><TELOPT_OLD_ENVIRON>\n");
			EnvNewOld = 1;
		} else {
			SendToServerOne(IAC);
			SendToServerOne(WONT);
			SendToServerOne(TELOPT_OLD_ENVIRON);
			DebugSndPrint("<IAC><WONT><TELOPT_OLD_ENVIRON>\n");
		}
		break;
	case TELOPT_NEW_ENVIRON:	/* New - Environment Variables */
		DebugRcvPrint("[TELOPT_NEW_ENVIRON]\n");
		if (EnvNewOld == -1) {
			SendToServerOne(IAC);
			SendToServerOne(WILL);
			SendToServerOne(TELOPT_NEW_ENVIRON);
			DebugSndPrint("<IAC><WILL><TELOPT_NEW_ENVIRON>\n");
			EnvNewOld = 2;
		} else {
			SendToServerOne(IAC);
			SendToServerOne(WONT);
			SendToServerOne(TELOPT_NEW_ENVIRON);
			DebugSndPrint("<IAC><WONT><TELOPT_NEW_ENVIRON>\n");
		}
		break;
	case TELOPT_XDISPLOC:	/* X Display Location */
		DebugRcvPrint("[TELOPT_XDISPLOC]\n");
		SendToServerOne(IAC);
		SendToServerOne(WILL);
		SendToServerOne(TELOPT_XDISPLOC);
		DebugSndPrint("<IAC><WILL><TELOPT_XDISPLOC>\n");
		break;
	default:
		DebugRcvPrint("[%03d]\n", (c & 0xFF));
		SendToServerOne(IAC);
		SendToServerOne(WONT);
		SendToServerOne(c);
		DebugSndPrint("<IAC><WONT><%03d>\n", (c & 0xFF));
		break;
	}
	cRecvState = 0;
	return 0;
}

/* WILLR}h */
int IacWill(char c)
{
	switch ((int)(c & 0xFF)) {
	case TELOPT_ECHO:	/* Echo */
		DebugRcvPrint("[TELOPT_ECHO]\n");
		SendToServerOne(IAC);
		SendToServerOne(DO);
		SendToServerOne(TELOPT_ECHO);
		DebugSndPrint("<IAC><DO><TELOPT_ECHO>\n");
		Echo = 0;
		break;
	case TELOPT_SGA:	/* Suppress go ahead */
		DebugRcvPrint("[TELOPT_SGA]\n");
		SendToServerOne(IAC);
		SendToServerOne(DO);
		SendToServerOne(TELOPT_SGA);
		DebugSndPrint("<IAC><DO><TELOPT_SGA>\n");
		break;
	default:
		DebugRcvPrint("[%03d]\n", (c & 0xFF));
		SendToServerOne(IAC);
		SendToServerOne(DONT);
		SendToServerOne(c);
		DebugSndPrint("<IAC><DONT><%03d>\n", (c & 0xFF));
		break;
	}
	cRecvState = 0;
	return 0;
}

/* DONTR}h */
int IacDont(char c)
{
	switch ((int)(c & 0xFF)) {
	case TELOPT_ECHO:	/* Echo */
		DebugRcvPrint("[TELOPT_ECHO]\n");
		SendToServerOne(IAC);
		SendToServerOne(WONT);
		SendToServerOne(TELOPT_ECHO);
		DebugSndPrint("<IAC><WONT><TELOPT_ECHO>\n");
		Echo = 0;
		break;
	default:
		DebugRcvPrint("[%03d]\n", (c & 0xFF));
		SendToServerOne(IAC);
		SendToServerOne(WONT);
		SendToServerOne(c);
		DebugSndPrint("<IAC><WONT><%03d>\n", (c & 0xFF));
		break;
	}
	cRecvState = 0;
	return 0;
}

/* WONTR}h */
int IacWont(char c)
{
	switch ((int)(c & 0xFF)) {
	case TELOPT_ECHO:	/* Echo */
		DebugRcvPrint("[TELOPT_ECHO]\n");
		SendToServerOne(IAC);
		SendToServerOne(DONT);
		SendToServerOne(TELOPT_ECHO);
		DebugSndPrint("<IAC><DONT><TELOPT_ECHO>\n");
		Echo = 1;
		break;
	default:
		DebugRcvPrint("[%03d]\n", (c & 0xFF));
		SendToServerOne(IAC);
		SendToServerOne(DONT);
		SendToServerOne(c);
		DebugSndPrint("<IAC><DONT><%03d>\n", (c & 0xFF));
		break;
	}
	cRecvState = 0;
	return 0;
}

/* TuR}h */
int IacSb(char c)
{
	if (cCtrlState[1] == 0) {
		switch ((int)(c & 0xFF)) {
		case TELOPT_TTYPE:	/* Terminal Type */
			DebugRcvPrint("[TELOPT_TTYPE]");
			cCtrlState[1] = c;
			cCtrlState[2] = 3;
			break;
		case TELOPT_XDISPLOC:	/* X Display Location */
			DebugRcvPrint("[TELOPT_XDISPLOC]");
			cCtrlState[1] = c;
			cCtrlState[2] = 3;
			break;
		case TELOPT_NAWS:	/* Window Size */
			DebugRcvPrint("[TELOPT_NAWS]");
			cCtrlState[1] = c;
			cCtrlState[2] = 3;
			break;
		case TELOPT_OLD_ENVIRON:	/* Old - Environment Variables */
			DebugRcvPrint("[TELOPT_OLD_ENVIRON]");
			cCtrlState[1] = c;
			cCtrlState[2] = 3;
			break;
		case TELOPT_NEW_ENVIRON:	/* New - Environment Variables */
			DebugRcvPrint("[TELOPT_NEW_ENVIRON]");
			cCtrlState[1] = c;
			cCtrlState[2] = 3;
			break;
		default:
			DebugRcvPrint("[%03d]", (c & 0xFF));
			cCtrlState[1] = c;
			cCtrlState[2] = 3;
			break;
		}
		return 0;
	}
	switch ((int)(cCtrlState[1] & 0xFF)) {
	case TELOPT_TTYPE:	/* Terminal Type */
		cCtrlState[2]--;
		switch (cCtrlState[2]) {
		case 2:
			DebugRcvPrint("[%03d]", (c & 0xFF));/* Must by TELQUAL_SEND */
			break;
		case 1:
			DebugRcvPrint("[IAC]");/* Must by IAC */
			break;
		default:
			DebugRcvPrint("[SE]\n");/* Must by SE */
			SendTTYPE();
			cRecvState = 0;
			break;
		}
		break;
	case TELOPT_XDISPLOC:	/* X Display Location */
		cCtrlState[2]--;
		switch (cCtrlState[2]) {
		case 2:
			DebugRcvPrint("[%03d]", (c & 0xFF));/* Must by TELQUAL_SEND */
			break;
		case 1:
			DebugRcvPrint("[IAC]");/* Must by IAC */
			break;
		default:
			DebugRcvPrint("[SE]\n");/* Must by SE */
			SendXDISPLOC();
			cRecvState = 0;
			break;
		}
		break;
	case TELOPT_NAWS:	/* Window Size */
		cCtrlState[2]--;
		switch (cCtrlState[2]) {
		case 2:
			DebugRcvPrint("[%03d]", (c & 0xFF));/* Must by TELQUAL_SEND */
			break;
		case 1:
			DebugRcvPrint("[IAC]");/* Must by IAC */
			break;
		default:
			DebugRcvPrint("[SE]\n");/* Must by SE */
			SendNAWS();
			cRecvState = 0;
			break;
		}
		break;
	case TELOPT_OLD_ENVIRON:	/* Old - Environment Variables */
		switch (cCtrlState[2]) {
		case 3:
			DebugRcvPrint("[%03d]", (c & 0xFF));/* Must by TELQUAL_SEND */
			cCtrlState[2]--;
			break;
		case 2:
			if ((int)(c & 0xFF) == IAC) {
				DebugRcvPrint("[IAC]");
				cCtrlState[2]--;
			} else {
				DebugRcvPrint("[%03d]", (c & 0xFF));
			}
			break;
		default:
			DebugRcvPrint("[SE]\n");/* Must by SE */
			SendOLD_ENVIRON();
			cRecvState = 0;
			break;
		}
		break;
	case TELOPT_NEW_ENVIRON:	/* New - Environment Variables */
		switch (cCtrlState[2]) {
		case 3:
			DebugRcvPrint("[%03d]", (c & 0xFF));/* Must by TELQUAL_SEND */
			cCtrlState[2]--;
			break;
		case 2:
			if ((int)(c & 0xFF) == IAC) {
				DebugRcvPrint("[IAC]");
				cCtrlState[2]--;
			} else {
				DebugRcvPrint("[%03d]", (c & 0xFF));
			}
			break;
		default:
			DebugRcvPrint("[SE]\n");/* Must by SE */
			SendNEW_ENVIRON();
			cRecvState = 0;
			break;
		}
		break;
	default:
		cCtrlState[2]--;
		switch (cCtrlState[2]) {
		case 2:
			DebugRcvPrint("[%03d]", (c & 0xFF));/* Must by TELQUAL_SEND */
			break;
		case 1:
			DebugRcvPrint("[IAC]");/* Must by IAC */
			break;
		default:
			DebugRcvPrint("[SE]\n");/* Must by SE */
			SendDefaultSb(c);
			cRecvState = 0;
			break;
		}
		break;
	}
	return 0;
}

/* ^[~i^CvM */
int SendTTYPE(void)
{
	SendToServerOne(IAC);
	SendToServerOne(SB);
	SendToServerOne(TELOPT_TTYPE);
	SendToServerOne(TELQUAL_IS);
	SendToServerPrint(szTerminalType);
	DebugSndPrint("<IAC><SB><TELOPT_TTYPE><TELQUAL_IS><%s>", szTerminalType);

	SendToServerOne(IAC);
	SendToServerOne(SE);
	DebugSndPrint("<IAC><SE>\n");

	return 0;
}

/* XfBXvCM */
int SendXDISPLOC(void)
{
	char *ptr;

	SendToServerOne(IAC);
	SendToServerOne(SB);
	SendToServerOne(TELOPT_XDISPLOC);
	SendToServerOne(TELQUAL_IS);
	if ((ptr = getenv("DISPLAY")) != NULL) {
		SendToServerPrint(ptr);
		DebugSndPrint("<IAC><SB><TELOPT_XDISPLOC><TELQUAL_IS><%s>", ptr);
	} else {
		SendToServerPrint(":0.0");
		DebugSndPrint("<IAC><SB><TELOPT_XDISPLOC><TELQUAL_IS><%s>", ":0.0");
	}
	SendToServerOne(IAC);
	SendToServerOne(SE);
	DebugSndPrint("<IAC><SE>\n");

	return 0;
}

/* ʃTCYM */
int SendNAWS(void)
{
	short rows, cols;

	SendToServerOne(IAC);
	SendToServerOne(SB);
	SendToServerOne(TELOPT_NAWS);
	GetTerminalWindowSize(&rows, &cols);
	SendToServerShort(cols);
	SendToServerShort(rows);
	SendToServerOne(IAC);
	SendToServerOne(SE);
	DebugSndPrint("<IAC><SB><TELOPT_NAWS><%d,%d><IAC><SE>\n", cols, rows);

	return 0;
}

/* ^CvM */
int SendOLD_ENVIRON(void)
{
	SendToServerOne(IAC);
	SendToServerOne(SB);
	SendToServerOne(TELOPT_OLD_ENVIRON);
	SendToServerOne(TELQUAL_IS);
	DebugSndPrint("<IAC><SB><TELOPT_OLD_ENVIRON><TELQUAL_IS>");
	SendToServerOne(IAC);
	SendToServerOne(SE);
	DebugSndPrint("<IAC><SE>\n");

	return 0;
}

/* V^CvM */
int SendNEW_ENVIRON(void)
{
	SendToServerOne(IAC);
	SendToServerOne(SB);
	SendToServerOne(TELOPT_NEW_ENVIRON);
	SendToServerOne(TELQUAL_IS);
	DebugSndPrint("<IAC><SB><TELOPT_NEW_ENVIRON><TELQUAL_IS>");
	SendToServerOne(IAC);
	SendToServerOne(SE);
	DebugSndPrint("<IAC><SE>\n");

	return 0;
}

/* TuR}hftHgM */
int SendDefaultSb(char c)
{
	SendToServerOne(IAC);
	SendToServerOne(SB);
	SendToServerOne(c);
	SendToServerOne(TELQUAL_IS);
	DebugSndPrint("<IAC><SB><%03d><TELQUAL_IS>", (c & 0xFF));
	SendToServerOne(IAC);
	SendToServerOne(SE);
	DebugSndPrint("<IAC><SE>\n");

	return 0;
}
