/*-----------------------------------------------------
   VT100.C -- Telnet Client, Terminal Emulator Program
             vt100[
   Copyright (c) 2002-2006
        Hiroto Shigeta. All rights reserved.
  -----------------------------------------------------*/
#include <windows.h>
#include "monitor.h"
#include "socket.h"
#include "mycolor.h"

/* Variables in CODESET.C */
extern BOOL bJisKanji;

/* Variables in TELNET.C */
extern char cRecvState;
extern char cCtrlState[32];

/* \ */
#define ATTR_BOLD      0x01
#define ATTR_UNDERLINE 0x02
#define ATTR_BLINK     0x04
#define ATTR_REVERSE   0x08

/* [R[h */
int VT100Control(char c)
{
	int  i;
	char *p;

	if (cCtrlState[0] == 0) {
		switch (c) {
		case '[':
			DebugRcvPrint("[[]");
			cCtrlState[0] = c;
			break;
		case 'D':
			/* ʂ1sփXN[ */
			DebugRcvPrint("[D]\n");
			ScrollDownMon();
			cRecvState = 0;
			break;
		case 'O':
			DebugRcvPrint("[O]");
			cCtrlState[0] = c;
			break;
		case 'M':
			/* ʂ1sփXN[ */
			DebugRcvPrint("[M]\n");
			ScrollUpMon();
			cRecvState = 0;
			break;
		case '7':
			/* LbgʒuL */
			DebugRcvPrint("[7]\n");
			SaveCaret();
			cRecvState = 0;
			break;
		case '8':
			/* Lbgʒu𕜌 */
			DebugRcvPrint("[8]\n");
			LoadCaret();
			cRecvState = 0;
			break;
		case '$':
			DebugRcvPrint("[$]");
			cCtrlState[0] = c;
			break;
		case '(':
			DebugRcvPrint("[(]");
			cCtrlState[0] = c;
			break;
		default:
			DebugRcvPrint("[%03d]\n", (c & 0xFF));
			cRecvState = 0;
			break;
		}
	} else if (cCtrlState[0] == '[') {
		if (cCtrlState[1] == 0) {
			switch (c) {
			case ';':
				DebugRcvPrint("[;]");
				cCtrlState[1] = c;
				break;
			case 'A':
				/* Lbgֈړ */
				DebugRcvPrint("[A]\n");
				MoveCaretUpOne();
				cRecvState = 0;
				break;
			case 'B':
				/* Lbgֈړ */
				DebugRcvPrint("[B]\n");
				MoveCaretDownOne();
				cRecvState = 0;
				break;
			case 'C':
				/* LbgEֈړ */
				DebugRcvPrint("[C]\n");
				MoveCaretRightOne();
				cRecvState = 0;
				break;
			case 'D':
				/* Lbgֈړ */
				DebugRcvPrint("[D]\n");
				MoveCaretLeftOne();
				cRecvState = 0;
				break;
			case 'J':
				/* Lbg肠Ƃ̕\NA */
				DebugRcvPrint("[J]\n");
				ClearMonAfterCaret();
				cRecvState = 0;
				break;
			case 'K':
				/* Lbg̈ʒus̏I܂ŃNA */
				DebugRcvPrint("[K]\n");
				ClearLineAfterCaret();
				cRecvState = 0;
				break;
			case '0': case '1': case '2': case '3': case '4': case '5':
			case '6': case '7': case '8': case '9':
				DebugRcvPrint("[%c]", c);
				cCtrlState[1] = c;
				break;
			case 'm':
				/* ׂĂ̑Itɂ */
				DebugRcvPrint("[m]");
				ClearAttr();
				cRecvState = 0;
				break;
			case '?':
				DebugRcvPrint("[?]");
				cCtrlState[1] = c;
				break;
			case 'H':
				/* Lbgz[|WVɖ߂ */
				DebugRcvPrint("[H]\n");
				MoveCaret(0, 0);
				cRecvState = 0;
				break;
			default:
				DebugRcvPrint("[%03d]\n", (c & 0xFF));
				cRecvState = 0;
				break;
			}
		} else if (cCtrlState[1] == ';') {
			if (cCtrlState[2] == 0) {
				switch (c) {
				case 'H':
					DebugRcvPrint("[H]");/* Must by H */
					cCtrlState[2] = 'H';
					break;
				default:/* Ô */
					DebugRcvPrint("[%c]\n", c);
					cRecvState = 0;
					break;
				}
			} else if (cCtrlState[3] == 0) {
				switch (c) {
				case '\x1B':
					DebugRcvPrint("[ESC]");/* Must by ESC */
					break;
				default:
					DebugRcvPrint("[%03d]", (c & 0xFF));
					break;
				}
				cCtrlState[3] = '\x1B';
			} else if (cCtrlState[4] == 0) {
				DebugRcvPrint("[%c]", c);/* Must by [ */
				cCtrlState[4] = '[';
			} else if (cCtrlState[5] == 0) {
				DebugRcvPrint("[%c]", c);/* Must by 2 */
				cCtrlState[5] = '2';
			} else {
				/* ʂNAăLbgz[|WVɖ߂ */
				DebugRcvPrint("[%c]\n", c);/* Must by J */
				ClearMon();
				cRecvState = 0;
			}
		} else if (cCtrlState[1] == '?') {
			if (cCtrlState[2] == 0) {
				DebugRcvPrint("[%c]", c);
				cCtrlState[2] = c;
			} else if (cCtrlState[3] == 0) {
				switch (cCtrlState[2]) {
				case '1':
					DebugRcvPrint("[%c]", c);/* h or l */
					cCtrlState[3] = c;
					break;
				case '2':
					DebugRcvPrint("[%c]", c);/* Must by 5 */
					cCtrlState[3] = '5';
					break;
				case '7':
					/* }[W@\I܂̓Itɂ */
					DebugRcvPrint("[%c]\n", c);/* h or l */
					cRecvState = 0;
					break;
				default:
					DebugRcvPrint("[%c]\n", c);/* h or l */
					cRecvState = 0;
					break;
				}
			} else if (cCtrlState[4] == 0) {
				switch (cCtrlState[2]) {
				case '1':
					switch (c) {
					case '\x1B':
						DebugRcvPrint("[ESC]");/* Must by ESC */
						break;
					default:
						DebugRcvPrint("[%03d]", (c & 0xFF));
						break;
					}
					cCtrlState[4] = '\x1B';
					break;
				case '2':
					/* Lbgʏ̖邳ɂ */
					DebugRcvPrint("[%c]\n", c);/* Must by h */
					cRecvState = 0;
				}
			} else {
				/* L[pbhI܂̓Itɂ */
				DebugRcvPrint("[%c]\n", c);/* = or > */
				cRecvState = 0;
			}
		} else {
			switch (c) {
			case 'm':
				DebugRcvPrint("[m]");
				switch (atoi(&cCtrlState[1])) {
				case 1:
					/* {[hIɂ */
					AddAttr(ATTR_BOLD);
					break;
				case 4:
					/* A_[C[hJn */
					AddAttr(ATTR_UNDERLINE);
					break;
				case 5:
					/* _ŕ\Iɂ */
					AddAttr(ATTR_BLINK);
					break;
				case 7:
					/* ]\Iɂ */
					AddAttr(ATTR_REVERSE);
					break;
				default:
					/* ׂĂ̑Itɂ */
					ClearAttr();
					break;
				}
				if ((p = strchr(&cCtrlState[1], ';')) != NULL) {
					switch (atoi(p + 1)) {
					case 1:
						/* {[hIɂ */
						AddAttr(ATTR_BOLD);
						break;
					case 4:
						/* A_[C[hJn */
						AddAttr(ATTR_UNDERLINE);
						break;
					case 5:
						/* _ŕ\Iɂ */
						AddAttr(ATTR_BLINK);
						break;
					case 7:
						/* ]\Iɂ */
						AddAttr(ATTR_REVERSE);
						break;
					}
				}
				cRecvState = 0;
				break;
			case 'A':
				/* Lbgֈړ */
				DebugRcvPrint("[A]\n");
				i = atoi(&cCtrlState[1]);
				while (i > 0) {
					MoveCaretUpOne();
					i--;
				}
				cRecvState = 0;
				break;
			case 'B':
				/* Lbgֈړ */
				DebugRcvPrint("[B]\n");
				i = atoi(&cCtrlState[1]);
				while (i > 0) {
					MoveCaretDownOne();
					i--;
				}
				cRecvState = 0;
				break;
			case 'C':
				/* LbgEֈړ */
				DebugRcvPrint("[C]\n");
				i = atoi(&cCtrlState[1]);
				while (i > 0) {
					MoveCaretRightOne();
					i--;
				}
				cRecvState = 0;
				break;
			case 'D':
				/* Lbgֈړ */
				DebugRcvPrint("[D]\n");
				i = atoi(&cCtrlState[1]);
				while (i > 0) {
					MoveCaretLeftOne();
					i--;
				}
				cRecvState = 0;
				break;
			case 'H':
				/* LbgWwňړ */
				DebugRcvPrint("[H]\n");
				MoveCaret(atoi(strchr(cCtrlState, ';') + 1) - 1,
						  atoi(&cCtrlState[1]) - 1);
				cRecvState = 0;
				break;
			case 'r':
				/* XN[͈͂̐ݒ */
				DebugRcvPrint("[r]\n");
				SetScrollRangeMon(atoi(&cCtrlState[1]) - 1,
								  atoi(strchr(cCtrlState, ';') + 1) - 1);
				cRecvState = 0;
				break;
			case '0': case '1': case '2': case '3': case '4': case '5':
			case '6': case '7': case '8': case '9': case ';':
				DebugRcvPrint("[%c]", c);
				for (i = 0; cCtrlState[i] != 0; i++);
				if (i < sizeof(cCtrlState)) {
					cCtrlState[i] = c;
				}
				break;
			case 'J':
				DebugRcvPrint("[J]\n");
				switch (cCtrlState[1]) {
				case '1': case '2':
					/* Lbg荶̕\NA */
					ClearMonBeforeCaret();
					if (cCtrlState[1] == '1') {
						break;
					}/* '2'̂Ƃ͑ONA(X[) */
				default:
					/* Lbg肠Ƃ̕\NA */
					ClearMonAfterCaret();
					break;
				}
				cRecvState = 0;
				break;
			case 'K':
				DebugRcvPrint("[K]\n");
				switch (cCtrlState[1]) {
				case '1': case '2':
					/* s̐擪Lbg̈ʒu܂łNA */
					ClearLineBeforeCaret();
					if (cCtrlState[1] == '1') {
						break;
					}/* '2'̂Ƃ͑ONA(X[) */
				default:
					/* Lbg̈ʒus̏I܂łNA */
					ClearLineAfterCaret();
					break;
				}
				cRecvState = 0;
				break;
			case 'L':
				/* LbgʒủɎw肳ꂽ̍sǉ */
				DebugRcvPrint("[L]\n");
				i = atoi(&cCtrlState[1]);
				AddLine(i);
				cRecvState = 0;
				break;
			case 'M':
				/* Lbgs牺̍sw肳ꂽ폜 */
				DebugRcvPrint("[M]\n");
				i = atoi(&cCtrlState[1]);
				DeleteLine(i);
				cRecvState = 0;
				break;
			case 'P':
				/* Lbg̈ʒuw肳ꂽ̕폜 */
				DebugRcvPrint("[P]\n");
				i = atoi(&cCtrlState[1]);
				DeleteChar(i);
				cRecvState = 0;
				break;
			default:
				DebugRcvPrint("[%03d]\n", (c & 0xFF));
				cRecvState = 0;
				break;
			}
		}
	} else if (cCtrlState[0] == 'O') {
		switch (c) {
		case 'A':
			/* Lbgֈړ */
			DebugRcvPrint("[A]\n");
			MoveCaretUpOne();
			cRecvState = 0;
			break;
		case 'B':
			/* Lbgֈړ */
			DebugRcvPrint("[B]\n");
			MoveCaretDownOne();
			cRecvState = 0;
			break;
		case 'C':
			/* LbgEֈړ */
			DebugRcvPrint("[C]\n");
			MoveCaretRightOne();
			cRecvState = 0;
			break;
		case 'D':
			/* Lbgֈړ */
			DebugRcvPrint("[D]\n");
			MoveCaretLeftOne();
			cRecvState = 0;
			break;
		default:
			DebugRcvPrint("[%03d]\n", (c & 0xFF));
			cRecvState = 0;
			break;
		}
	} else if (cCtrlState[0] == '$') {
		/* JISin */
		DebugRcvPrint("[%c]", c);/* @ or B */
		bJisKanji = TRUE;
		cRecvState = 0;
	} else if (cCtrlState[0] == '(') {
		/* JISout */
		DebugRcvPrint("[%c]", c);/* B or J */
		bJisKanji = FALSE;
		cRecvState = 0;
	} else {/* Ô */
		DebugRcvPrint("[%03d]\n", (c & 0xFF));
		cRecvState = 0;
	}
	return 0;
}

/* L[M */
BOOL VT100KeyDown(WPARAM wParam, LPARAM lParam)
{
	switch (wParam) {
	case VK_UP:
		SendToServerOne('\x1B');
		SendToServerOne('O');
		SendToServerOne('A');
		return TRUE;
	case VK_DOWN:
		SendToServerOne('\x1B');
		SendToServerOne('O');
		SendToServerOne('B');
		return TRUE;
	case VK_RIGHT:
		SendToServerOne('\x1B');
		SendToServerOne('O');
		SendToServerOne('C');
		return TRUE;
	case VK_LEFT:
		SendToServerOne('\x1B');
		SendToServerOne('O');
		SendToServerOne('D');
		return TRUE;
	case VK_F1:
		SendToServerOne('\x1B');
		SendToServerOne('O');
		SendToServerOne('P');
		return TRUE;
	case VK_F2:
		SendToServerOne('\x1B');
		SendToServerOne('O');
		SendToServerOne('Q');
		return TRUE;
	case VK_F3:
		SendToServerOne('\x1B');
		SendToServerOne('O');
		SendToServerOne('R');
		return TRUE;
	case VK_F4:
		SendToServerOne('\x1B');
		SendToServerOne('O');
		SendToServerOne('S');
		return TRUE;
	}
	return FALSE;
}

static HFONT hULFont = NULL;	/* Under Line */
static HFONT hBSFont = NULL;	/* Bold Style */
static HFONT hBUFont = NULL;	/* Bold Style Under Line */
static int   iExtra1 = 0;		/* ʏtHgƂ̍̕(p) */
static int   iExtra2 = 0;		/* ʏtHgƂ̍̕(Sp) */

/* eLXgʏo */
BOOL TextOutVT100(HDC hdc, int nXStart, int nYStart, LPCTSTR lpString, int cbString,
				  HFONT hFont, int cxChar, BOOL bSelected, char *pchAttr)
{
	COLORREF crTextColor, crBkColor;
	HFONT    hOldFont;
	int      i;
	char     chAttrBefore = pchAttr[0];
	char     *pStartPos   = (char *)lpString;

	crTextColor = GetMyColor(bSelected ? COLOR_HIGHLIGHTTEXT : COLOR_WINDOWTEXT);
	crBkColor   = GetMyColor(bSelected ? COLOR_HIGHLIGHT : COLOR_WINDOW);
	for (i = 0; i <= cbString; i++) {
		if (i < cbString && chAttrBefore == pchAttr[i]) {
			continue;
		}
		if (chAttrBefore & ATTR_UNDERLINE && chAttrBefore & ATTR_BOLD) {
			hOldFont = SelectObject(hdc, hBUFont);
		} else if (chAttrBefore & ATTR_UNDERLINE) {
			hOldFont = SelectObject(hdc, hULFont);
		} else if (chAttrBefore & ATTR_BOLD) {
			hOldFont = SelectObject(hdc, hBSFont);
		}
		SetTextColor(hdc, chAttrBefore & ATTR_REVERSE ? crBkColor : crTextColor);
		if (bSelected || chAttrBefore & ATTR_REVERSE) {
			SetBkMode(hdc, OPAQUE);
			SetBkColor(hdc, chAttrBefore & ATTR_REVERSE ? crTextColor : crBkColor);
		} else {
			SetBkMode(hdc, TRANSPARENT);
		}
		AdjustTextOut(hdc, nXStart + (pStartPos - lpString) * cxChar, nYStart, pStartPos,
					  i - (pStartPos - lpString),
					  chAttrBefore & ATTR_BOLD ? iExtra1 : 0,
					  chAttrBefore & ATTR_BOLD ? iExtra2 : 0);
		if (chAttrBefore & ATTR_UNDERLINE || chAttrBefore & ATTR_BOLD) {
			SelectObject(hdc, hOldFont);
		}
		if (i < cbString) {
			chAttrBefore = pchAttr[i];
			pStartPos    = (char *)&lpString[i];
		}
	}
	return TRUE;
}

/* tHgύX */
void ChangeFontVT100(HWND hwnd, HFONT hFont)
{
	LOGFONT    logfont;
	HDC        hdc;
	HFONT      hOldFont;
	TEXTMETRIC tm;
	SIZE       size1, size2, size3, size4;

	GetObject(hFont, sizeof(LOGFONT), (PVOID)&logfont);
	if (hULFont != NULL) {
		DeleteObject(hULFont);
	}
	logfont.lfUnderline = TRUE;
	hULFont = CreateFontIndirect(&logfont);
	if (hBUFont != NULL) {
		DeleteObject(hBUFont);
	}
	logfont.lfWeight = FW_BOLD;		/* ʏtHgȂωȂ */
	hBUFont = CreateFontIndirect(&logfont);
	if (hBSFont != NULL) {
		DeleteObject(hBSFont);
	}
	logfont.lfUnderline = FALSE;
	hBSFont = CreateFontIndirect(&logfont);

	hdc = GetDC(hwnd);
	hOldFont = SelectObject(hdc, hFont);
	GetTextExtentPoint32(hdc, "ab", 2, &size1);
	GetTextMetrics(hdc, &tm);
	if (tm.tmCharSet == SHIFTJIS_CHARSET) {
		GetTextExtentPoint32(hdc, "", 4, &size2);
	}
	SelectObject(hdc, hBSFont);
	GetTextExtentPoint32(hdc, "ab", 2, &size3);
	iExtra1 = (size3.cx - size1.cx) / 2;
	if (tm.tmCharSet == SHIFTJIS_CHARSET) {
		GetTextExtentPoint32(hdc, "", 4, &size4);
		iExtra2 = (size4.cx - size2.cx) / 2;
	} else {
		iExtra2 = iExtra1;
	}
	SelectObject(hdc, hOldFont);
	ReleaseDC(hwnd, hdc);
}
